from copy import deepcopy
from typing import Union, List, Dict, TYPE_CHECKING, Optional

from ravendb.primitives import constants
from ravendb.documents.commands.batches import SingleNodeBatchCommand, ClusterWideBatchCommand, CommandType
from ravendb.documents.operations.patch import PatchStatus
from ravendb.documents.session.event_args import AfterSaveChangesEventArgs
from ravendb.documents.session.misc import TransactionMode
from ravendb.documents.session.document_info import DocumentInfo
from ravendb.exceptions.raven_exceptions import ClientVersionMismatchException
from ravendb.json.result import BatchCommandResult
from ravendb.tools.utils import CaseInsensitiveDict

if TYPE_CHECKING:
    from ravendb.documents.session.document_session_operations.in_memory_document_session_operations import (
        InMemoryDocumentSessionOperations,
    )


class BatchOperation:
    def __init__(self, session: "InMemoryDocumentSessionOperations"):
        self._session = session
        self._entities: List[object] = []
        self._session_commands_count: Union[None, int] = None
        self._all_commands_count: Union[None, int] = None
        self._on_successful_request: Union[
            None, "InMemoryDocumentSessionOperations.SaveChangesData.ActionsToRunOnSuccess"
        ] = None
        self._modifications: Union[None, Dict[str, DocumentInfo]] = None

    def create_request(self) -> Union[None, SingleNodeBatchCommand]:
        result = self._session.prepare_for_save_changes()
        self._on_successful_request = result.on_success
        self._session_commands_count = len(result.session_commands)
        result.session_commands.extend(result.deferred_commands)

        self._session.validate_cluster_transaction(result)

        self._all_commands_count = len(result.session_commands)

        if self._all_commands_count == 0:
            return None

        self._session.increment_requests_count()

        self._entities = result.entities

        if self._session.transaction_mode == TransactionMode.CLUSTER_WIDE:
            return ClusterWideBatchCommand(
                self._session.conventions,
                result.session_commands,
                result.options,
                self._session.disable_atomic_document_writes_in_cluster_wide_transaction,
            )
        return SingleNodeBatchCommand(self._session.conventions, result.session_commands, result.options)

    def set_result(self, result: BatchCommandResult) -> None:
        def get_command_type(obj_node: dict) -> CommandType:
            c_type = obj_node.get("Type")
            if not c_type:
                return CommandType.NONE

            type_as_str = str(c_type)

            _command_type = CommandType.from_csharp_value_str(type_as_str)
            return _command_type

        if result.results is None:
            self._throw_on_null_result()
            return

        self._on_successful_request.clear_session_state_after_successful_save_changes()

        if self._session.transaction_mode == TransactionMode.CLUSTER_WIDE:
            if result.transaction_index <= 0:
                raise ClientVersionMismatchException(
                    "Cluster transaction was send to a node that is not supporting "
                    "it. So it was executed ONLY on the requested node on " + self._session.request_executor.url
                )

        for i in range(self._session_commands_count):
            batch_result = result.results[i]
            if batch_result is None:
                continue

            command_type = get_command_type(batch_result)

            if command_type == CommandType.PUT:
                self._handle_put(i, batch_result, False)
            elif command_type == CommandType.FORCE_REVISION_CREATION:
                self._handle_force_revision_creation(batch_result)
            elif command_type == CommandType.DELETE:
                self._handle_delete(batch_result)
            elif command_type == CommandType.COMPARE_EXCHANGE_PUT:
                self._handle_compare_exchange_put(batch_result)
            elif command_type == CommandType.COMPARE_EXCHANGE_DELETE:
                self._handle_compare_exchange_delete(batch_result)
            else:
                raise ValueError(f"Command {command_type} is not supported")

        for i in range(self._session_commands_count, self._all_commands_count):
            batch_result = result.results[i]
            if batch_result is None:
                continue

            command_type = get_command_type(batch_result)
            if command_type == CommandType.PUT:
                self._handle_put(i, batch_result, False)
            elif command_type == CommandType.DELETE:
                self._handle_delete(batch_result)
            elif command_type == CommandType.PATCH:
                self._handle_patch(batch_result)
            elif command_type == CommandType.ATTACHMENT_PUT:
                self._handle_attachment_put(batch_result)
            elif command_type == CommandType.ATTACHMENT_DELETE:
                self._handle_attachment_delete(batch_result)
            elif command_type == CommandType.ATTACHMENT_MOVE:
                self._handle_attachment_move(batch_result)
            elif command_type == CommandType.ATTACHMENT_COPY:
                self._handle_attachment_copy(batch_result)
            elif command_type in (
                CommandType.COMPARE_EXCHANGE_PUT,
                CommandType.COMPARE_EXCHANGE_DELETE,
                CommandType.FORCE_REVISION_CREATION,
            ):
                pass
            elif command_type == CommandType.COUNTERS:
                self._handle_counters(batch_result)
            elif command_type == CommandType.TIME_SERIES:
                break  # todo: RavenDB-13474 add to time series cache
            elif command_type == CommandType.TIME_SERIES_COPY or command_type == CommandType.BATCH_PATCH:
                break
            else:
                raise ValueError(f"Command {command_type} is not supported")

        self._finalize_result()

    def _finalize_result(self):
        if not self._modifications:
            return

        for key, document_info in self._modifications.items():
            self._apply_metadata_modifications(key, document_info)

    def _apply_metadata_modifications(self, key: str, document_info: DocumentInfo):
        document_info.metadata_instance = None
        document_info.metadata = deepcopy(document_info.metadata)
        document_info.metadata[constants.Documents.Metadata.CHANGE_VECTOR] = document_info.change_vector
        document_copy = deepcopy(document_info.document)
        document_copy[constants.Documents.Metadata.KEY] = document_info.metadata

        document_info.document = document_copy

    def _get_or_add_modifications(
        self, key: str, document_info: DocumentInfo, apply_modifications: bool
    ) -> DocumentInfo:
        if not self._modifications:
            self._modifications = CaseInsensitiveDict()

        modified_document_info = self._modifications.get(key)
        if modified_document_info is not None:
            if apply_modifications:
                self._apply_metadata_modifications(key, modified_document_info)
        else:
            self._modifications[key] = modified_document_info = document_info

        return modified_document_info

    def _handle_compare_exchange_put(self, batch_result: dict) -> None:
        self._handle_compare_exchange_internal(CommandType.COMPARE_EXCHANGE_PUT, batch_result)

    def _handle_compare_exchange_delete(self, batch_result: dict) -> None:
        self._handle_compare_exchange_internal(CommandType.COMPARE_EXCHANGE_DELETE, batch_result)

    def _handle_compare_exchange_internal(self, command_type: CommandType, batch_result: dict) -> None:
        key: str = batch_result.get("Key")
        if not key:
            self._throw_missing_field(command_type, "Key")

        index: int = batch_result.get("Index")
        if not index:
            self._throw_missing_field(command_type, "Index")

        cluster_session = self._session.cluster_transaction
        cluster_session.update_state(key, index)

    def _handle_patch(self, batch_result: dict) -> None:
        patch_status = batch_result.get("PatchStatus")
        if not patch_status:
            self._throw_missing_field(CommandType.PATCH, "PatchStatus")

        status = PatchStatus(patch_status)
        if status == PatchStatus.CREATED or PatchStatus.PATCHED:
            document = batch_result.get("ModifiedDocument")
            if not document:
                return

            key = self._get_string_field(batch_result, CommandType.PUT, "Id")
            session_document_info = self._session.documents_by_id.get(key)
            if session_document_info is None:
                return

            document_info = self._get_or_add_modifications(key, session_document_info, True)

            change_vector = self._get_string_field(batch_result, CommandType.PATCH, "ChangeVector")
            last_modified = self._get_string_field(batch_result, CommandType.PATCH, "LastModified")

            document_info.change_vector = change_vector
            document_info.metadata[constants.Documents.Metadata.KEY] = key
            document_info.metadata[constants.Documents.Metadata.CHANGE_VECTOR] = change_vector
            document_info.metadata[constants.Documents.Metadata.LAST_MODIFIED] = last_modified

            document_info.document = document
            self._apply_metadata_modifications(key, document_info)

            if document_info.entity is not None:
                self._session.entity_to_json.populate_entity(document_info.entity, key, document_info.document)
                self._session.after_save_changes_invoke(
                    AfterSaveChangesEventArgs(self._session, document_info.key, document_info.entity)
                )

    def _handle_delete(self, batch_result: dict) -> None:
        self._handle_delete_internal(batch_result, CommandType.DELETE)

    def _handle_delete_internal(self, batch_result: dict, command_type: CommandType):
        key = self._get_string_field(batch_result, command_type, "Id")
        document_info = self._session.documents_by_id.get(key)
        if document_info is None:
            return

        self._session.documents_by_id.pop(key, None)

        if document_info.entity is not None:
            self._session.documents_by_entity.pop(document_info.entity, None)
            self._session.deleted_entities.discard(document_info.entity)

    def _handle_force_revision_creation(self, batch_result: dict) -> None:
        # When forcing a revision for a document that does Not have any revisions yet then the HasRevisions flag is added to the document.
        # In this case we need to update the tracked entities in the session with the document new change-vector.

        if not self._get_boolean_field(batch_result, CommandType.FORCE_REVISION_CREATION, "RevisionCreated"):
            # no forced revision was created...nothing to update
            return

        id_ = self._get_string_field(batch_result, CommandType.FORCE_REVISION_CREATION, constants.Documents.Metadata.ID)
        change_vector = self._get_string_field(
            batch_result, CommandType.FORCE_REVISION_CREATION, constants.Documents.Metadata.CHANGE_VECTOR
        )

        document_info = self._session.documents_by_id.get(id_)
        if document_info is None:
            return

        document_info.change_vector = change_vector

        self._handle_metadata_modifications(document_info, batch_result, id_, change_vector)

        after_save_changes_event_args = AfterSaveChangesEventArgs(
            self._session, document_info.key, document_info.entity
        )
        self._session.after_save_changes_invoke(after_save_changes_event_args)

    def _handle_put(self, index: int, batch_result: dict, is_deferred: bool) -> None:
        entity = None
        document_info = None

        if not is_deferred:
            entity = self._entities[index]
            document_info = self._session.documents_by_entity.get(entity)
            if document_info is None:
                return

        key = self._get_string_field(batch_result, CommandType.PUT, constants.Documents.Metadata.ID)
        change_vector = self._get_string_field(
            batch_result, CommandType.PUT, constants.Documents.Metadata.CHANGE_VECTOR
        )

        if is_deferred:
            session_document_info = self._session.documents_by_id.get(key)
            if session_document_info is None:
                return

            document_info = self._get_or_add_modifications(key, session_document_info, True)
            entity = document_info.entity

        self._handle_metadata_modifications(document_info, batch_result, key, change_vector)

        self._session.documents_by_id.update({document_info.key: document_info})

        if entity:
            self._session.generate_entity_id_on_the_client.try_set_identity(entity, key)

        self._session.after_save_changes_invoke(
            AfterSaveChangesEventArgs(self._session, document_info.key, document_info.entity)
        )

    def _handle_metadata_modifications(
        self, document_info: DocumentInfo, batch_result: dict, key: str, change_vector: str
    ) -> None:
        for property_name, value in batch_result.items():
            if "Type" == property_name:
                continue

            document_info.metadata[property_name] = value

        document_info.key = key
        document_info.change_vector = change_vector

        self._apply_metadata_modifications(key, document_info)

    def _handle_counters(self, batch_result: Dict) -> None:
        doc_id = self._get_string_field(batch_result, CommandType.COUNTERS, "Id")
        counters_detail: Optional[dict] = batch_result.get("CountersDetail", None)
        if counters_detail is None:
            self._throw_missing_field(CommandType.COUNTERS, "CountersDetail")

        counters: Optional[Dict] = counters_detail.get("Counters", None)
        if counters is None:
            self._throw_missing_field(CommandType.COUNTERS, "Counters")

        cache = self._session.counters_by_doc_id.get(doc_id, None)
        if cache is None:
            cache = [False, CaseInsensitiveDict()]
            self._session.counters_by_doc_id[doc_id] = cache

        change_vector = self._get_string_field(batch_result, CommandType.COUNTERS, "DocumentChangeVector", False)
        if change_vector is not None:
            document_info = self._session.documents_by_id.get(doc_id, None)
            if document_info is not None:
                document_info.change_vector = change_vector

        for counter in counters:
            name = counter.get("CounterName", None)
            value = counter.get("TotalValue", None)

            if name is not None and value is not None:
                cache[1][name] = value

    def _handle_attachment_put(self, batch_result: dict) -> None:
        self._handle_attachment_put_internal(
            batch_result, CommandType.ATTACHMENT_PUT, "Id", "Name", "DocumentChangeVector"
        )

    def _handle_attachment_copy(self, batch_result: dict) -> None:
        self._handle_attachment_put_internal(
            batch_result, CommandType.ATTACHMENT_COPY, "Id", "Name", "DocumentChangeVector"
        )

    def _handle_attachment_move(self, batch_result: dict) -> None:
        self._handle_attachment_delete_internal(
            batch_result, CommandType.ATTACHMENT_MOVE, "Id", "Name", "DocumentChangeVector"
        )
        self._handle_attachment_put_internal(
            batch_result, CommandType.ATTACHMENT_MOVE, "DestinationId", "DestinationName", "DocumentChangeVector"
        )

    def _handle_attachment_delete(self, batch_result: dict) -> None:
        self._handle_attachment_delete_internal(
            batch_result, CommandType.ATTACHMENT_DELETE, constants.Documents.Metadata.ID, "Name", "DocumentChangeVector"
        )

    def _handle_attachment_delete_internal(
        self,
        batch_result: dict,
        command_type: CommandType,
        id_field_name: str,
        attachment_name_field_name: str,
        document_change_vector_field_name: str,
    ) -> None:
        key = self._get_string_field(batch_result, command_type, id_field_name)

        session_document_info = self._session.documents_by_id.get_value(key)
        if session_document_info is None:
            return

        document_info = self._get_or_add_modifications(key, session_document_info, True)

        document_change_vector = self._get_string_field(
            batch_result, command_type, document_change_vector_field_name, False
        )
        if document_change_vector:
            document_info.change_vector = document_change_vector

        attachments_json = document_info.metadata.get(constants.Documents.Metadata.ATTACHMENTS)
        if not attachments_json:
            return

        name = self._get_string_field(batch_result, command_type, attachment_name_field_name)

        attachments = []
        document_info.metadata[constants.Documents.Metadata.ATTACHMENTS] = attachments

        for attachment in attachments_json:
            attachment_name = self._get_string_field(attachment, command_type, "Name")
            if attachment_name == name:
                continue

            attachments.append(attachment)

    def _handle_attachment_put_internal(
        self,
        batch_result: dict,
        command_type: CommandType,
        id_field_name: str,
        attachment_name_field_name: str,
        document_change_vector_field_name: str,
    ) -> None:
        key = self._get_string_field(batch_result, command_type, id_field_name)

        session_document_info = self._session.documents_by_id.get_value(key)
        if session_document_info is None:
            return

        document_info = self._get_or_add_modifications(key, session_document_info, False)
        document_change_vector = self._get_string_field(
            batch_result, command_type, document_change_vector_field_name, False
        )
        if document_change_vector:
            document_info.change_vector = document_change_vector

        attachments = document_info.metadata.get(constants.Documents.Metadata.ATTACHMENTS)
        if attachments is None:
            attachments = []
            document_info.metadata[constants.Documents.Metadata.ATTACHMENTS] = attachments

        dynamic_node = {
            "ChangeVector": self._get_string_field(batch_result, command_type, "ChangeVector"),
            "ContentType": self._get_string_field(batch_result, command_type, "ContentType"),
            "Hash": self._get_string_field(batch_result, command_type, "Hash"),
            "Name": self._get_string_field(batch_result, command_type, "Name"),
            "Size": self._get_string_field(batch_result, command_type, "Size"),
        }
        attachments.append(dynamic_node)

    def _get_string_field(
        self, json: dict, command_type: CommandType, field_name: str, throw_on_missing: Optional[bool] = True
    ) -> str:
        json_node = json.get(field_name, None)
        if throw_on_missing and json_node is None:
            self._throw_missing_field(command_type, field_name)
        return str(json_node)

    def _get_int_field(self, json: dict, command_type: CommandType, field_name: str) -> int:
        json_node = json.get(field_name)
        if not isinstance(json_node, int):
            self._throw_missing_field(command_type, field_name)
        return json_node

    def _get_boolean_field(self, json: dict, command_type: CommandType, field_name: str) -> bool:
        json_node = json.get(field_name, None)
        if not isinstance(json_node, bool):
            self._throw_missing_field(command_type, field_name)
        return json_node

    def _throw_on_null_result(self) -> None:
        raise ValueError(
            "Reveived empty response from the server. This is not supposed to happend and is likely a bug."
        )

    def _throw_missing_field(self, c_type: CommandType, field_name: str) -> None:
        raise ValueError(f"{c_type} response is invalid. Field '{field_name}' is missing.")



import json
from typing import Optional

import requests

from ravendb.http.server_node import ServerNode
from ravendb.http.raven_command import RavenCommand, VoidRavenCommand


class GetNextOperationIdCommand(RavenCommand[int]):
    def __init__(self):
        super(GetNextOperationIdCommand, self).__init__(int)
        self._node_tag = 0

    @property
    def node_tag(self):
        return self._node_tag

    def is_read_request(self) -> bool:
        return False  # disable caching

    def create_request(self, node: ServerNode) -> requests.Request:
        return requests.Request("GET", f"{node.url}/databases/{node.database}/operations/next-operation-id")

    def set_response(self, response: Optional[str], from_cache: bool) -> None:
        json_node = json.loads(response)
        self.result = json_node.get("Id", None)
        self._node_tag = json_node.get("NodeTag", None)


class KillOperationCommand(VoidRavenCommand):
    def __init__(self, operation_id: int, node_tag: Optional[str] = None):
        super(KillOperationCommand, self).__init__()
        self._id = operation_id
        self._selected_node_tag = node_tag

    def create_request(self, node: ServerNode) -> requests.Request:
        return requests.Request("POST", f"{node.url}/databases/{node.database}/operations/kill?id={self._id}")



from __future__ import annotations
from typing import Generic, TypeVar, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from ravendb.documents.session.query import DocumentQuery

_T = TypeVar("_T")


class GroupByField:
    def __init__(self, field_name: Optional[str] = None, projected_name: Optional[str] = None):
        self.field_name = field_name
        self.projected_name = projected_name


class GroupByDocumentQuery(Generic[_T]):
    def __init__(self, query: DocumentQuery[_T]):
        self.__query = query

    def select_key(
        self, field_name: Optional[str] = None, projected_name: Optional[str] = None
    ) -> GroupByDocumentQuery[_T]:
        self.__query._group_by_key(field_name, projected_name)
        return self

    def select_sum(self, field: GroupByField, *fields: GroupByField) -> DocumentQuery[_T]:
        if field is None:
            raise ValueError("field cannot be None")

        self.__query._group_by_sum(field.field_name, field.projected_name)

        if not fields:
            return self.__query

        for f in fields:
            self.__query._group_by_sum(f.field_name, f.projected_name)

        return self.__query

    def select_count(self, projected_name: str = "count") -> DocumentQuery[_T]:
        self.__query._group_by_count(projected_name)
        return self.__query



from typing import Optional, Dict


class InvalidOperationException(Exception):
    pass


class NonRecoverableSubscriptionException(Exception):
    pass


class ErrorResponseException(Exception):
    pass


class DocumentDoesNotExistsException(Exception):
    pass


class NonUniqueObjectException(Exception):
    pass


class FetchConcurrencyException(Exception):
    pass


class ArgumentOutOfRangeException(Exception):
    pass


class DatabaseDoesNotExistException(Exception):
    pass


class AuthorizationException(Exception):
    pass


class IndexDoesNotExistException(Exception):
    pass


class TimeoutException(Exception):
    pass


class AuthenticationException(Exception):
    pass


class AllTopologyNodesDownException(Exception):
    pass


class UnsuccessfulRequestException(Exception):
    pass


class RequestedNodeUnavailableException(Exception):
    pass


class AggregateException(Exception):
    pass


class NotSupportedException(Exception):
    pass


class AggregateException(Exception):
    pass


class ChangeProcessingException(Exception):
    pass


class ChangeProcessingException(Exception):
    pass


# <---------- Subscription Exceptions ---------->


class SubscriptionException(Exception):
    pass


class SubscriptionInUseException(SubscriptionException):
    pass


class SubscriptionClosedException(SubscriptionException):
    def __init__(self, *args, can_reconnect: Optional[bool] = None):
        super(SubscriptionClosedException, self).__init__(*args)
        self.can_reconnect = can_reconnect


class SubscriptionInvalidStateException(SubscriptionException):
    pass


class SubscriptionDoesNotExistException(SubscriptionException):
    pass


class SubscriptionDoesNotBelongToNodeException(SubscriptionException):
    def __init__(
        self,
        *args,
        appropriate_node: Optional[str] = None,
        reasons: Dict[str, str] = None,
        register_connection_duration_in_ticks: bool = None,
    ):
        super(SubscriptionDoesNotBelongToNodeException, self).__init__(*args)
        self.appropriate_node = appropriate_node
        self.reasons = reasons
        self.register_connection_duration_in_ticks = register_connection_duration_in_ticks


class SubscriptionChangeVectorUpdateConcurrencyException(SubscriptionException):
    pass


class SubscriberErrorException(Exception):
    pass


class InvalidNetworkTopologyException(Exception):
    pass


class SubscriptionMessageTypeException(Exception):
    pass


class RavenTimeoutException(Exception):
    pass



from __future__ import annotations

import json
from typing import Optional, List, TYPE_CHECKING, Dict

import requests

from ravendb import ServerNode
from ravendb.documents.operations.definitions import MaintenanceOperation, VoidMaintenanceOperation
from ravendb.documents.operations.ongoing_tasks import OngoingTaskType
from ravendb.http.raven_command import RavenCommand, VoidRavenCommand
from ravendb.http.topology import RaftCommand
from ravendb.documents.operations.backups.settings import PeriodicBackupConfiguration
from ravendb.serverwide.operations.common import ServerOperation, T, VoidServerOperation, DatabaseSettings
from ravendb.serverwide.operations.ongoing_tasks import IServerWideTask, ServerWideTaskResponse
from ravendb.tools.utils import Utils
from ravendb.util.util import RaftIdGenerator


if TYPE_CHECKING:
    from ravendb.documents.conventions import DocumentConventions


class ServerWideBackupConfiguration(PeriodicBackupConfiguration, IServerWideTask):
    def __init__(self, name_prefix: str = "Server Wide Backup"):
        super(ServerWideBackupConfiguration, self).__init__()
        IServerWideTask.__init__(self)

        self._name_prefix = name_prefix


class PutServerWideBackupConfigurationOperation(
    ServerOperation["PutServerWideBackupConfigurationOperation.PutServerWideBackupConfigurationResponse"]
):
    def __init__(self, configuration: ServerWideBackupConfiguration):
        if configuration is None:
            raise ValueError("Configuration cannot be None")

        self._configuration = configuration

    def get_command(self, conventions: "DocumentConventions") -> RavenCommand[T]:
        return self.PutServerWideBackupConfigurationCommand(self._configuration)

    class PutServerWideBackupConfigurationCommand(
        RavenCommand["PutServerWideBackupConfigurationOperation.PutServerWideBackupConfigurationResponse"], RaftCommand
    ):
        def __init__(self, configuration: ServerWideBackupConfiguration):
            super().__init__(PutServerWideBackupConfigurationOperation.PutServerWideBackupConfigurationResponse)
            if configuration is None:
                raise ValueError("Configuration cannot be None")
            self._configuration = configuration

        def is_read_request(self) -> bool:
            return False

        def get_raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()

        def create_request(self, node: ServerNode) -> requests.Request:
            url = f"{node.url}/admin/configuration/server-wide/backup"
            request = requests.Request(method="PUT", url=url)
            request.data = self._configuration.to_json()
            return request

        def set_response(self, response: Optional[str], from_cache: bool) -> None:
            self.result = PutServerWideBackupConfigurationOperation.PutServerWideBackupConfigurationResponse.from_json(
                json.loads(response)
            )

    class PutServerWideBackupConfigurationResponse(ServerWideTaskResponse):
        pass


class GetServerWideBackupConfigurationsOperation(ServerOperation[List[ServerWideBackupConfiguration]]):
    def get_command(self, conventions: "DocumentConventions") -> RavenCommand[List[ServerWideBackupConfiguration]]:
        return self.GetServerWideBackupConfigurationsCommand()

    class GetServerWideBackupConfigurationsCommand(RavenCommand[List[ServerWideBackupConfiguration]]):
        def __init__(self):
            super().__init__(ServerWideBackupConfiguration)

        def is_read_request(self) -> bool:
            return True

        def create_request(self, node: ServerNode) -> requests.Request:
            url = f"{node.url}/admin/configuration/server-wide/tasks?type=Backup"
            return requests.Request(method="GET", url=url)

        def set_response(self, response: Optional[str], from_cache: bool) -> None:
            if response is None:
                return

            self.result = [
                ServerWideBackupConfiguration.from_json(result_json) for result_json in json.loads(response)["Results"]
            ]


class GetServerWideBackupConfigurationOperation(ServerOperation[ServerWideBackupConfiguration]):
    def __init__(self, name: str):
        if name is None:
            raise ValueError("Name cannot be None")

        self._name = name

    def get_command(self, conventions: "DocumentConventions") -> RavenCommand[T]:
        return self.GetServerWideBackupConfigurationCommand(self._name)

    class GetServerWideBackupConfigurationCommand(RavenCommand[ServerWideBackupConfiguration]):
        def __init__(self, name: str):
            super().__init__(ServerWideBackupConfiguration)

            if name is None:
                raise ValueError("Name cannot be None")

            self._name = name

        def is_read_request(self) -> bool:
            return True

        def create_request(self, node: ServerNode) -> requests.Request:
            url = f"{node.url}/admin/configuration/server-wide/tasks?type=Backup&name=" + Utils.quote_key(self._name)
            return requests.Request(method="GET", url=url)

        def set_response(self, response: Optional[str], from_cache: bool) -> None:
            if response is None:
                return

            results = [ServerWideBackupConfiguration.from_json(bc_json) for bc_json in json.loads(response)["Results"]]
            if not results:
                return

            if len(results) > 1:
                self._throw_invalid_response()

            self.result = results[0]


class DeleteServerWideTaskOperation(VoidServerOperation):
    def __init__(self, name: str, type_: OngoingTaskType):
        if name is None:
            raise ValueError("Name cannot be None")

        self._name = name
        self._type = type_

    def get_command(self, conventions: "DocumentConventions") -> "VoidRavenCommand":
        return self.DeleteServerWideTaskCommand(self._name, self._type)

    class DeleteServerWideTaskCommand(VoidRavenCommand, RaftCommand):
        def __init__(self, name: str, type_: OngoingTaskType):
            super().__init__()
            if name is None:
                raise ValueError("Name cannot be None")

            self._name = name
            self._type = type_

        def is_read_request(self) -> bool:
            return False

        def get_raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()

        def create_request(self, node: ServerNode) -> requests.Request:
            url = f"{node.url}/admin/configuration/server-wide/task?type={self._type.value}&name={Utils.quote_key(self._name)}"
            return requests.Request(method="DELETE", url=url)


class GetDatabaseSettingsOperation(MaintenanceOperation[DatabaseSettings]):
    def __init__(self, database_name: str = None):
        if database_name is None:
            raise ValueError("Database name cannot be None")
        self._database_name = database_name

    def get_command(self, conventions: "DocumentConventions") -> "RavenCommand[DatabaseSettings]":
        return self.GetDatabaseSettingsCommand(self._database_name)

    class GetDatabaseSettingsCommand(RavenCommand[DatabaseSettings]):
        def __init__(self, database_name: str = None):
            super().__init__(DatabaseSettings)

            if database_name is None:
                raise ValueError("Database name cannot be None")

            self._database_name = database_name

        def is_read_request(self) -> bool:
            return False

        def create_request(self, node: ServerNode) -> requests.Request:
            url = f"{node.url}/databases/{self._database_name}/admin/record"
            return requests.Request(method="GET", url=url)

        def set_response(self, response: Optional[str], from_cache: bool) -> None:
            if response is None:
                self.result = None
                return

            self.result = DatabaseSettings.from_json(json.loads(response))


class PutDatabaseSettingsOperation(VoidMaintenanceOperation):
    def __init__(self, database_name: str, configuration_settings: Dict[str, str]):
        if database_name is None:
            raise ValueError("Database name cannot be None")
        self._database_name = database_name

        if configuration_settings is None:
            raise ValueError("Configuration settings cannot be None")
        self._configuration_settings = configuration_settings

    def get_command(self, conventions: "DocumentConventions") -> "VoidRavenCommand":
        return self.PutDatabaseSettingsCommand(self._configuration_settings, self._database_name)

    class PutDatabaseSettingsCommand(VoidRavenCommand, RaftCommand):
        def __init__(self, configuration_settings: Dict[str, str], database_name: str):
            super().__init__()
            if database_name is None:
                raise ValueError("Database name cannot be None.")
            self._database_name = database_name

            if configuration_settings is None:
                raise ValueError("Configuration settings cannot be None.")
            self._configuration_settings = configuration_settings

        def is_read_request(self) -> bool:
            return False

        def get_raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()

        def create_request(self, node: ServerNode) -> requests.Request:
            url = f"{node.url}/databases/{self._database_name}/admin/configuration/settings"

            return requests.Request(method="PUT", url=url, data=self._configuration_settings)



from __future__ import annotations

import datetime
import json
from abc import abstractmethod
from enum import Enum
from typing import Callable, Union, Optional, TYPE_CHECKING, List, Set, Dict

import requests

from ravendb.documents.operations.counters import (
    CounterOperation,
    DocumentCountersOperation,
    CounterOperationType,
)
from ravendb.documents.operations.time_series import TimeSeriesOperation
from ravendb.documents.session.misc import TransactionMode, ForceRevisionStrategy
from ravendb.http.raven_command import RavenCommand
from ravendb.http.server_node import ServerNode
from ravendb.json.result import BatchCommandResult
from ravendb.tools.utils import CaseInsensitiveSet, Utils
from ravendb.util.util import RaftIdGenerator

if TYPE_CHECKING:
    from ravendb.documents.conventions import DocumentConventions
    from ravendb.documents.operations.patch import PatchRequest
    from ravendb.documents.session.document_session_operations.in_memory_document_session_operations import (
        InMemoryDocumentSessionOperations,
    )


class CommandType(Enum):
    NONE = "None"
    PUT = "PUT"
    PATCH = "PATCH"
    DELETE = "DELETE"
    ATTACHMENT_PUT = "AttachmentPUT"
    ATTACHMENT_DELETE = "AttachmentDELETE"
    ATTACHMENT_MOVE = "AttachmentMOVE"
    ATTACHMENT_COPY = "AttachmentCOPY"
    COMPARE_EXCHANGE_PUT = "CompareExchangePUT"
    COMPARE_EXCHANGE_DELETE = "CompareExchangeDELETE"

    FORCE_REVISION_CREATION = "ForceRevisionCreation"

    COUNTERS = "Counters"
    TIME_SERIES = "TimeSeries"
    TIME_SERIES_BULK_INSERT = "TIME_SERIES_BULK_INSERT"
    TIME_SERIES_COPY = "TIME_SERIES_COPY"
    BATCH_PATCH = "BatchPATCH"
    CLIENT_ANY_COMMAND = "CLIENT_ANY_COMMAND"
    CLIENT_MODIFY_DOCUMENT_COMMAND = "CLIENT_MODIFY_DOCUMENT_COMMAND"

    def __str__(self):
        return self.value

    @classmethod
    def from_csharp_value_str(cls, value: str) -> CommandType:
        if value == "None":
            return cls.NONE
        elif value == "PUT":
            return cls.PUT
        elif value == "PATCH":
            return cls.PATCH
        elif value == "DELETE":
            return cls.DELETE
        elif value == "AttachmentPUT":
            return cls.ATTACHMENT_PUT
        elif value == "AttachmentDELETE":
            return cls.ATTACHMENT_DELETE
        elif value == "AttachmentMOVE":
            return cls.ATTACHMENT_MOVE
        elif value == "AttachmentCOPY":
            return cls.ATTACHMENT_COPY
        elif value == "CompareExchangePUT":
            return cls.COMPARE_EXCHANGE_PUT
        elif value == "CompareExchangeDELETE":
            return cls.COMPARE_EXCHANGE_DELETE
        elif value == "Counters":
            return cls.COUNTERS
        elif value == "BatchPATCH":
            return cls.BATCH_PATCH
        elif value == "ForceRevisionCreation":
            return cls.FORCE_REVISION_CREATION
        elif value == "TimeSeries":
            return cls.TIME_SERIES
        else:
            raise ValueError(f"Unable to parse: {value}")


# --------------COMMAND--------------
class SingleNodeBatchCommand(RavenCommand):
    def __init__(
        self,
        conventions: DocumentConventions,
        commands: List[CommandData],
        options: BatchOptions = None,
        mode: TransactionMode = TransactionMode.SINGLE_NODE,
    ):
        if not conventions:
            raise ValueError("Conventions cannot be None")
        if commands is None:
            raise ValueError("Commands cannot be None")
        for command in commands:
            if command is None:
                raise ValueError("Command cannot be None")

        super().__init__(result_class=BatchCommandResult)
        self.__conventions = conventions
        self.__commands = commands
        self.__options = options
        self.__mode = mode
        self.__attachment_streams: List[bytes] = list()

        for command in commands:
            if isinstance(command, PutAttachmentCommandData):
                if self.__attachment_streams is None:
                    self.__attachment_streams = []
                stream = command.stream
                if stream in self.__attachment_streams:
                    raise RuntimeError(
                        "It is forbidden to re-use the same stream for more than one attachment. "
                        "Use a unique stream per put attachment command."
                    )
                self.__attachment_streams.append(stream)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    @property
    def is_read_request(self):
        return False

    def create_request(self, node: ServerNode) -> requests.Request:
        request = requests.Request(method="POST")
        files = {"main": None}
        for command in self.__commands:
            if command.command_type == CommandType.ATTACHMENT_PUT:
                command: PutAttachmentCommandData
                files[command.name] = (
                    command.name,
                    command.stream,
                    command.content_type,
                    {"Command-Type": "AttachmentStream"},
                )
            if not request.data:
                request.data = {"Commands": []}
            request.data["Commands"].append(command.serialize(self.__conventions))

        if self.__mode == TransactionMode.CLUSTER_WIDE:
            request.data["TransactionMode"] = "ClusterWide"
            if files:
                request.use_stream = True

        if len(files) > 1:
            files["main"] = json.dumps(request.data, default=self.__conventions.json_default_method)
            request.files = files
            request.data = None

        sb = [f"{node.url}/databases/{node.database}/bulk_docs?"]
        self._append_options(sb)

        request.url = "".join(sb)

        return request

    def _append_options(self, sb: List[str]) -> None:
        if self.__options is None:
            return
        sb.append("?")
        replication_options = self.__options.replication_options
        if replication_options:
            sb.append(
                f"&waitForReplicasTimeout=" f"{Utils.timedelta_to_str(replication_options.wait_for_replicas_timeout)}"
            )
            sb.append(
                f"&throwOnTimeoutInWaitForReplicas="
                f"{'true' if replication_options.throw_on_timeout_in_wait_for_replicas else 'false'}"
            )
            sb.append(
                f"&numberOfReplicasToWaitFor="
                f"{'majority' if replication_options.majority else replication_options.number_of_replicas_to_wait_for}"
            )

        index_options = self.__options.index_options
        if index_options:
            sb.append(f"&waitForIndexesTimeout={Utils.timedelta_to_str(index_options.wait_for_indexes_timeout)}")
            sb.append(f"&waitForIndexThrow={'true' if index_options.throw_on_timeout_in_wait_for_indexes else 'false'}")
            if index_options.wait_for_specific_indexes:
                for specific_index in index_options.wait_for_specific_indexes:
                    sb.append(f"&waitForSpecificIndex={specific_index.encode('utf-8')}")

    def set_response(self, response: str, from_cache: bool) -> None:
        if response is None:
            raise ValueError(
                "Got None response from the server after doing a batch, something is very wrong."
                " Probably a garbled response."
            )
        self.result = Utils.initialize_object(json.loads(response), self._result_class, True)


class ClusterWideBatchCommand(SingleNodeBatchCommand):
    def __init__(
        self,
        conventions: DocumentConventions,
        commands: List[CommandData],
        options: Optional[BatchOptions] = None,
        disable_atomic_documents_writes: Optional[bool] = None,
    ):
        super().__init__(conventions, commands, options, TransactionMode.CLUSTER_WIDE)

        self.__disable_atomic_document_writes = disable_atomic_documents_writes

    def get_raft_unique_request_id(self) -> str:
        return RaftIdGenerator.new_id()

    def _append_options(self, sb: List[str]) -> None:
        super()._append_options(sb)

        if self.__disable_atomic_document_writes is None:
            return

        sb.append("&disableAtomicDocumentWrites=")
        sb.append("true" if self.__disable_atomic_document_writes else "false")


# -------------- DATA ---------------


class CommandData:
    def __init__(
        self,
        key: str = None,
        name: str = None,
        change_vector: str = None,
        command_type: CommandType = None,
    ):
        self._key = key
        self._name = name
        self._change_vector = change_vector
        self._command_type = command_type

    @property
    def key(self) -> str:
        return self._key

    @property
    def name(self) -> str:
        return self._name

    @property
    def change_vector(self) -> str:
        return self._change_vector

    @property
    def command_type(self) -> CommandType:
        return self._command_type

    def on_before_save_changes(self, session: InMemoryDocumentSessionOperations) -> None:
        pass

    @abstractmethod
    def serialize(self, conventions: DocumentConventions) -> dict:
        pass


class DeleteCommandData(CommandData):
    def __init__(self, key: str, change_vector: str, original_change_vector: str = None):
        super(DeleteCommandData, self).__init__(key=key, command_type=CommandType.DELETE, change_vector=change_vector)
        self.original_change_vector = original_change_vector

    def serialize(self, conventions: DocumentConventions) -> dict:
        data = {"Id": self.key, "ChangeVector": self.change_vector, "Type": CommandType.DELETE}
        if self.original_change_vector is not None:
            data.update({"OriginalChangeVector": self.original_change_vector})
        return data


class PutCommandDataBase(CommandData):
    def __init__(
        self,
        key: str,
        change_vector: str,
        document: object,
        strategy: ForceRevisionStrategy = ForceRevisionStrategy.NONE,
    ):
        super(PutCommandDataBase, self).__init__(
            key=key, name=None, change_vector=change_vector, command_type=CommandType.PUT
        )
        self.__document = document
        self.__force_revisions_creation_strategy = strategy

    @property
    def document(self) -> object:
        return self.__document

    @property
    def force_revision_creation_strategy(self) -> ForceRevisionStrategy:
        return self.__force_revisions_creation_strategy

    def serialize(self, conventions: DocumentConventions) -> dict:
        result = {
            "Id": self._key,
            "ChangeVector": self._change_vector,
            "Document": self.__document,
            "Type": CommandType.PUT,
        }
        if self.force_revision_creation_strategy != ForceRevisionStrategy.NONE:
            result["ForceRevisionCreationStrategy"] = self.__force_revisions_creation_strategy
        return result


class PutCommandDataWithJson(PutCommandDataBase):
    def __init__(self, key, change_vector, document, strategy):
        super(PutCommandDataWithJson, self).__init__(key, change_vector, document, strategy)


class ForceRevisionCommandData(CommandData):
    def __init__(self, key: str):
        super(ForceRevisionCommandData, self).__init__(key=key, command_type=CommandType.FORCE_REVISION_CREATION)

    def serialize(self, conventions: DocumentConventions) -> dict:
        return {"Id": super().key, "Type": super().command_type}


class BatchPatchCommandData(CommandData):
    class IdAndChangeVector:
        def __init__(self, key: str, change_vector: str):
            self.key = key
            self.change_vector = change_vector

        @classmethod
        def create(cls, key: str, change_vector: str) -> BatchPatchCommandData.IdAndChangeVector:
            return cls(key, change_vector)

    def __init__(self, patch: PatchRequest, patch_if_missing: PatchRequest, *ids: Union[str, IdAndChangeVector]):
        if patch is None:
            raise ValueError("Patch cannot be None")

        super(BatchPatchCommandData, self).__init__(
            command_type=CommandType.BATCH_PATCH, name=None, change_vector=None, key=None
        )

        self.__seen_ids: Set[str] = CaseInsensitiveSet()
        self.__ids: List[BatchPatchCommandData.IdAndChangeVector] = []
        self.__patch = patch
        self.__patch_if_missing = patch_if_missing
        self.on_before_save_changes: Callable[[InMemoryDocumentSessionOperations], None] = None

        if not ids:
            raise ValueError("Ids cannot be None or empty collection")
        for key in ids:
            if isinstance(key, BatchPatchCommandData.IdAndChangeVector):
                self.__add(key.key, key.change_vector)
            elif isinstance(key, str):
                self.__add(key)
            raise TypeError(f"Unexpected type: {type(key)} of key")

    def __add(self, key: str, change_vector: Optional[str] = None):
        if not key:
            raise ValueError("Value cannot be None or whitespace")
        if not self.__seen_ids.add(key):
            raise ValueError(f"Could not add ID '{key}' because item with the same ID was already added")
        self.__ids.append(BatchPatchCommandData.IdAndChangeVector.create(key, change_vector))

    @property
    def ids(self):
        return self.__ids

    @property
    def patch(self):
        return self.__patch

    @property
    def patch_if_missing(self):
        return self.__patch_if_missing

    def serialize(self, conventions: DocumentConventions) -> dict:
        arr = []
        for kvp in self.__ids:
            if kvp.change_vector is not None:
                arr.append({"Id": kvp.key, "ChangeVector": kvp.change_vector})
            arr.append({"Id": kvp.key})
        if self.patch_if_missing is None:
            return {"Ids": arr, "Patch": self.patch.serialize(), "Type": "BatchPATCH"}
        return {
            "Ids": arr,
            "Patch": self.patch.serialize(),
            "Type": "BatchPATCH",
            "PatchIfMissing": self.patch_if_missing.serialize(),
        }


class PatchCommandData(CommandData):
    def __init__(
        self,
        key: str,
        change_vector: Union[None, str],
        patch: PatchRequest,
        patch_if_missing: Optional[PatchRequest] = None,
    ):
        super().__init__(key, None, change_vector, CommandType.PATCH)
        if not key:
            raise ValueError("Key cannot be None")

        if not patch:
            raise ValueError("Patch cannot be None")

        self.create_if_missing: Union[None, dict] = None
        self.__patch = patch
        self.__patch_if_missing = patch_if_missing
        self.return_document: Union[None, bool] = None

        def __consumer(session: InMemoryDocumentSessionOperations) -> None:
            self.return_document = session.advanced.is_loaded(key)

        self.__on_before_save_changes = __consumer

    @property
    def patch(self):
        return self.__patch

    @property
    def patch_if_missing(self):
        return self.__patch_if_missing

    @property
    def on_before_save_changes(self):
        return self.__on_before_save_changes

    def serialize(self, conventions: DocumentConventions) -> dict:
        data = {"Id": self.key, "ChangeVector": self.change_vector, "Patch": self.patch.serialize(), "Type": "PATCH"}
        if self.patch_if_missing:
            data.update({"PatchIfMissing": self.patch_if_missing.serialize()})
        if self.create_if_missing:
            data.update({"CreateIfMissing": self.create_if_missing})
        if self.return_document:
            data.update({"ReturnDocument": self.return_document})
        return data


class PutCompareExchangeCommandData(CommandData):
    def __init__(self, key: str, value: dict, index: int):
        super(PutCompareExchangeCommandData, self).__init__(key=key, command_type=CommandType.COMPARE_EXCHANGE_PUT)
        self.document = value
        self.index = index

    def serialize(self, conventions: DocumentConventions) -> dict:
        return {
            "Id": self.key,
            "Document": self.document,
            "Index": self.index,
            "Type": CommandType.COMPARE_EXCHANGE_PUT,
        }


class DeleteCompareExchangeCommandData(CommandData):
    def __init__(self, key: str, index: int):
        super(DeleteCompareExchangeCommandData, self).__init__(
            key=key, command_type=CommandType.COMPARE_EXCHANGE_DELETE
        )
        self.key = key
        self.index = index

    def serialize(self, conventions: DocumentConventions) -> dict:
        return {"Id": self.key, "Index": self.index, "Type": CommandType.COMPARE_EXCHANGE_DELETE}

    @property
    def key(self):
        return self._key

    @key.setter
    def key(self, value):
        self._key = value


class CountersBatchCommandData(CommandData):
    def __init__(
        self,
        document_id: str,
        counter_operations: Union[List[CounterOperation], CounterOperation],
        from_etl: Optional[bool] = None,
        change_vector=None,
    ):
        if not document_id:
            raise ValueError("None or empty document id is Invalid")
        if not counter_operations:
            raise ValueError("invalid counter_operations")
        if not isinstance(counter_operations, list):
            counter_operations = [counter_operations]

        super().__init__(key=document_id, command_type=CommandType.COUNTERS, change_vector=change_vector)
        self._from_etl = from_etl
        self._counters = DocumentCountersOperation(self.key, counter_operations)

    def has_increment(self, counter_name):
        self.has_operation_of_type(CounterOperationType.INCREMENT, counter_name)

    def has_delete(self, counter_name):
        self.has_operation_of_type(CounterOperationType.DELETE, counter_name)

    def has_operation_of_type(self, operation_type: CounterOperationType, counter_name: str):
        for op in self.counters.operations:
            if op.counter_name != counter_name:
                continue

            if op.counter_operation_type == operation_type:
                return True
        return False

    @property
    def counters(self):
        return self._counters

    def serialize(self, conventions: DocumentConventions) -> Dict:
        json_dict = {
            "Id": self.key,
            "Counters": self.counters.to_json(),
            "Type": self.command_type,
        }
        if self._from_etl:
            json_dict["FromEtl"] = self._from_etl
        return json_dict


class PutAttachmentCommandData(CommandData):
    def __init__(self, document_id: str, name: str, stream: bytes, content_type: str, change_vector: str):
        if not document_id:
            raise ValueError(document_id)
        if not name:
            raise ValueError(name)

        super(PutAttachmentCommandData, self).__init__(document_id, name, change_vector, CommandType.ATTACHMENT_PUT)
        self.__stream = stream
        self.__content_type = content_type

    @property
    def stream(self):
        return self.__stream

    @property
    def content_type(self):
        return self.__content_type

    def serialize(self, conventions: DocumentConventions) -> dict:
        return {
            "Id": self._key,
            "Name": self._name,
            "ContentType": self.__content_type,
            "ChangeVector": self._change_vector,
            "Type": str(self._command_type),
        }


class CopyAttachmentCommandData(CommandData):
    def __init__(
        self,
        source_document_id: str,
        source_name: str,
        destination_document_id: str,
        destination_name: str,
        change_vector: Optional[str],
    ):
        if source_document_id.isspace():
            raise ValueError("source_document_id is required")
        if source_name.isspace():
            raise ValueError("source_name is required")
        if destination_document_id.isspace():
            raise ValueError("destination_document_id is required")
        if destination_name.isspace():
            raise ValueError("destination_name is required")
        super().__init__(source_document_id, source_name, change_vector, CommandType.ATTACHMENT_COPY)
        self.destination_id = destination_document_id
        self.destination_name = destination_name

    def serialize(self, conventions: DocumentConventions) -> dict:
        return {
            "Id": self.key,
            "Name": self.name,
            "DestinationId": self.destination_id,
            "DestinationName": self.destination_name,
            "ChangeVector": self.change_vector,
            "Type": "AttachmentCOPY",
        }


class MoveAttachmentCommandData(CommandData):
    def __init__(
        self,
        key: str,
        name: str,
        destination_id: str,
        destination_name: str,
        change_vector: Optional[str],
    ):
        if key.isspace():
            raise ValueError("source_document_id is required")
        if name.isspace():
            raise ValueError("source_name is required")
        if destination_id.isspace():
            raise ValueError("destination_document_id is required")
        if destination_name.isspace():
            raise ValueError("destination_name is required")
        super().__init__(key, name, change_vector, CommandType.ATTACHMENT_MOVE)
        self.destination_id = destination_id
        self.destination_name = destination_name

    def serialize(self, conventions: DocumentConventions) -> dict:
        return {
            "Id": self.key,
            "Name": self.name,
            "DestinationId": self.destination_id,
            "DestinationName": self.destination_name,
            "ChangeVector": self.change_vector,
            "Type": "AttachmentMOVE",
        }


class DeleteAttachmentCommandData(CommandData):
    def __init__(self, document_id: str, name: str, change_vector: Optional[str]):
        if not document_id:
            raise ValueError(document_id)
        if not name:
            raise ValueError(name)
        super().__init__(document_id, name, change_vector, CommandType.ATTACHMENT_DELETE)

    def serialize(self, conventions: DocumentConventions) -> dict:
        return {"Id": self._key, "Name": self.name, "ChangeVector": self.change_vector, "Type": self.command_type}


class TimeSeriesBatchCommandData(CommandData):
    def __init__(
        self,
        document_id: str,
        name: str,
        appends: Optional[List[TimeSeriesOperation.AppendOperation]],
        deletes: Optional[List[TimeSeriesOperation.DeleteOperation]],
    ):
        if not document_id:
            raise ValueError("Document id cannot be None")
        if not name:
            raise ValueError("Name cannot be None")

        super(TimeSeriesBatchCommandData, self).__init__(document_id, name, None, CommandType.TIME_SERIES)
        self.time_series = TimeSeriesOperation()
        self.time_series.name = name

        if appends:
            for append_operation in appends:
                self.time_series.append(append_operation)
        if deletes:
            for delete_operation in deletes:
                self.time_series.delete(delete_operation)

    def serialize(self, conventions: DocumentConventions) -> dict:
        return {"Id": self.key, "TimeSeries": self.time_series.to_json(), "Type": "TimeSeries"}

    def on_before_save_changes(self, session: InMemoryDocumentSessionOperations) -> None:
        pass


# ------------ OPTIONS ------------


class ReplicationBatchOptions:
    def __init__(
        self,
        wait_for_replicas: bool = None,
        number_of_replicas_to_wait_for: int = None,
        wait_for_replicas_timeout: datetime.timedelta = None,
        majority: bool = None,
        throw_on_timeout_in_wait_for_replicas: bool = True,
    ):
        self.wait_for_replicas = wait_for_replicas
        self.number_of_replicas_to_wait_for = number_of_replicas_to_wait_for
        self.wait_for_replicas_timeout = wait_for_replicas_timeout
        self.majority = majority
        self.throw_on_timeout_in_wait_for_replicas = throw_on_timeout_in_wait_for_replicas


class IndexBatchOptions:
    def __init__(
        self,
        wait_for_indexes: bool = None,
        wait_for_indexes_timeout: datetime.timedelta = None,
        throw_on_timeout_in_wait_for_indexes: bool = None,
        wait_for_specific_indexes: List[str] = None,
    ):
        self.wait_for_indexes = wait_for_indexes
        self.wait_for_indexes_timeout = wait_for_indexes_timeout
        self.throw_on_timeout_in_wait_for_indexes = throw_on_timeout_in_wait_for_indexes
        self.wait_for_specific_indexes = wait_for_specific_indexes


class BatchOptions:
    def __init__(
        self,
        replication_options: Optional[ReplicationBatchOptions] = None,
        index_options: Optional[IndexBatchOptions] = None,
    ):
        self.replication_options = replication_options
        self.index_options = index_options



from __future__ import annotations

import datetime
import threading
import time
import uuid
from abc import abstractmethod
from concurrent.futures import ThreadPoolExecutor
from typing import TYPE_CHECKING, Optional
from typing import Union, List, Dict

from ravendb.exceptions.exceptions import (
    DatabaseDoesNotExistException,
    RequestedNodeUnavailableException,
    AllTopologyNodesDownException,
)
from ravendb.http.server_node import ServerNode

if TYPE_CHECKING:
    pass


class Topology:
    def __init__(self, etag: int, nodes: List[ServerNode]):
        self.etag = etag
        self.nodes = nodes


class ClusterTopology:
    def __init__(self):
        self.last_node_id = None
        self.topology_id = None
        self.etag = None

        self.members: Dict[str, str] = {}
        self.promotables: Dict[str, str] = {}
        self.watchers: Dict[str, str] = {}

    def __contains__(self, node: str) -> bool:
        if self.members is not None and node in self.members:
            return True
        if self.promotables is not None and node in self.promotables:
            return True

        return self.watchers is not None and node in self.watchers

    @classmethod
    def from_json(cls, json_dict: Dict) -> ClusterTopology:
        topology = cls()
        topology.etag = json_dict["Etag"]
        topology.topology_id = json_dict["TopologyId"]
        topology.members = json_dict["Members"]
        topology.promotables = json_dict["Promotables"]
        topology.watchers = json_dict["Watchers"]
        topology.last_node_id = json_dict["LastNodeId"]
        return topology


class UpdateTopologyParameters:
    def __init__(self, node: ServerNode):
        if not node:
            raise ValueError("Node cannot be None")
        self.node = node

        self.timeout_in_ms = 15_000
        self.force_update: Union[None, bool] = None
        self.debug_tag: Union[None, str] = None
        self.application_identifier: Union[None, uuid.UUID] = None


class NodeSelector:
    class __NodeSelectorState:
        def __init__(self, topology: Topology):
            self.topology = topology
            self.nodes = topology.nodes
            self.failures = [0] * len(topology.nodes)
            self.fastest_records = [0] * len(topology.nodes)
            self.fastest: Union[None, int] = None
            self.speed_test_mode = 0
            self.unlikely_everyone_faulted_choice_index: Optional[int] = 0

        @property
        def node_when_everyone_marked_as_faulted(self) -> CurrentIndexAndNode:
            index = self.unlikely_everyone_faulted_choice_index
            self.unlikely_everyone_faulted_choice_index = (self.unlikely_everyone_faulted_choice_index + 1) % len(
                self.nodes
            )

            return CurrentIndexAndNode(index, self.nodes[index])

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.__update_fastest_node_timer is not None:
            self.__update_fastest_node_timer.__exit__()

    def __init__(self, topology: Topology, thread_pool: ThreadPoolExecutor):
        self._state = self.__NodeSelectorState(topology)
        self.__update_fastest_node_timer: Union[None, time] = None
        self.__thread_pool_executor = thread_pool

    @property
    def topology(self) -> Topology:
        return self._state.topology

    def node_is_available(self, index: int) -> bool:
        return self._state.failures[index] == 0

    def on_failed_request(self, node_index: int) -> None:
        state = self._state
        if node_index < 0 or node_index >= len(state.failures):
            return
        state.failures[node_index] += 1

    def on_update_topology(self, topology: Topology, force_update: bool = False) -> bool:
        if topology is None:
            return False

        state_etag = self._state.topology.etag if self._state.topology.etag else 0
        topology_etag = topology.etag if topology.etag else 0

        if state_etag >= topology_etag and not force_update:
            return False

        state = NodeSelector.__NodeSelectorState(topology)
        self._state = state

        return True

    def get_requested_node(self, node_tag: str) -> CurrentIndexAndNode:
        state = self._state
        server_nodes = state.nodes
        for i in range(len(server_nodes)):
            if server_nodes[i].cluster_tag == node_tag:
                return CurrentIndexAndNode(i, server_nodes[i])

        if len(state.nodes) == 0:
            raise DatabaseDoesNotExistException("There are no nodes in the topology at all")
        raise RequestedNodeUnavailableException(f"Could not find requested node {node_tag}")

    def get_preferred_node(self) -> CurrentIndexAndNode:
        state = self._state
        return self.get_preferred_node_internal(state)

    @classmethod
    def get_preferred_node_internal(cls, state: NodeSelector.__NodeSelectorState) -> CurrentIndexAndNode:
        state_failures = state.failures
        server_nodes = state.nodes
        length = min(len(server_nodes), len(state_failures))
        for i in range(length):
            if state_failures[0] == 0:
                return CurrentIndexAndNode(i, server_nodes[i])
        return cls.unlikely_everyone_faulted_choice(state)

    def get_preferred_node_with_topology(self) -> CurrentIndexAndNodeAndEtag:
        state = self._state
        preferred_node = self.get_preferred_node_internal(state)
        etag = (state.topology.etag if state.topology.etag else -2) if state.topology else -2
        return CurrentIndexAndNodeAndEtag(preferred_node.current_index, preferred_node.current_node, etag)

    @staticmethod
    def unlikely_everyone_faulted_choice(state: NodeSelector.__NodeSelectorState) -> CurrentIndexAndNode:
        # if there are all marked as failed, we'll chose the next (the one in CurrentNodeIndex)
        # one so the user will get an error (or recover :-) )
        if len(state.nodes) == 0:
            raise DatabaseDoesNotExistException("There are no nodes in the topology at all")
        return state.node_when_everyone_marked_as_faulted

    def get_node_by_session_id(self, session_id: int) -> CurrentIndexAndNode:
        state = self._state
        if len(state.topology.nodes) == 0:
            raise AllTopologyNodesDownException("There are no nodes in the topology at all")
        index = abs(session_id % len(state.topology.nodes))
        for i in range(index, len(state.failures)):
            if state.failures[i] == 0 and state.nodes[i].server_role == ServerNode.Role.MEMBER:
                return CurrentIndexAndNode(i, state.nodes[i])

        for i in range(index):
            if state.failures[i] == 0 and state.nodes[i].server_role == ServerNode.Role.MEMBER:
                return CurrentIndexAndNode(i, state.nodes[i])

        return self.get_preferred_node()

    def get_fastest_node(self) -> CurrentIndexAndNode:
        state = self._state
        if state.failures[state.fastest] == 0 and state.nodes[state.fastest].server_role == ServerNode.Role.MEMBER:
            return CurrentIndexAndNode(state.fastest, state.nodes[state.fastest])

        # if the fastest node has failures, we'll immediately schedule
        # another run of finding who the fastest node is, in the mantime
        # we'll just use the server preferred node or failover as usual

        self._switch_to_speed_test_phase()
        return self.get_preferred_node()

    def restore_node_index(self, node_index: int) -> None:
        state = self._state
        if len(state.failures) <= node_index:
            return

        state.failures[node_index] = 0

    @staticmethod
    def _throw_empty_topology() -> None:
        raise RuntimeError("Empty database topology, this shouldn't happen.")

    def _switch_to_speed_test_phase(self) -> None:
        state = self._state

        if not state.speed_test_mode == 0:
            state.speed_test_mode = 1
            return

        state.fastest_records = len(state.fastest_records) * [0]

        state.speed_test_mode += 1

    @property
    def in_speed_test_phase(self) -> bool:
        return self._state.speed_test_mode > 1

    def record_fastest(self, index: int, node: ServerNode) -> None:
        state = self._state
        state_fastest = state.fastest_records

        # the following two checks are to verify that things didn't move
        # while we were computing the fastest node, we verify that the index
        # of the fastest node and rhe identity of the node didn't change during
        # our check

        if index < 0 or index >= len(state_fastest):
            return

        if node != state.nodes[index]:
            return

        state_fastest[index] += 1
        if state_fastest[index] >= 10:
            self.__select_fastest(state, index)

        state.speed_test_mode += 1

        if state.speed_test_mode <= len(state.nodes) * 10:
            return

        # too many concurrent speed tests are happening
        max_index = self.__find_max_index(state)
        self.__select_fastest(state, max_index)

    @staticmethod
    def __find_max_index(state: NodeSelector.__NodeSelectorState) -> int:
        state_fastest = state.fastest_records
        max_index = 0
        max_value = 0
        for i in range(len(state_fastest)):
            if max_value >= state_fastest[i]:
                continue

            max_index = i
            max_value = state_fastest[i]

        return max_index

    def __select_fastest(self, state: NodeSelector.__NodeSelectorState, index: int) -> None:
        state.fastest = index
        state.speed_test_mode = 0

        if self.__update_fastest_node_timer is not None:
            # todo: figure out how to reshedule timer, or put it in sleep for 60s
            threading.self.__update_fastest_node_timer

        else:
            self.__update_fastest_node_timer = threading.Timer(
                datetime.timedelta(minutes=1), self._switch_to_speed_test_phase
            )

    def schedule_speed_test(self) -> None:
        self._switch_to_speed_test_phase()


class CurrentIndexAndNode:
    def __init__(self, current_index: int, current_node: ServerNode):
        self.current_index = current_index
        self.current_node = current_node


class CurrentIndexAndNodeAndEtag:
    def __init__(self, current_index: int, current_node: ServerNode, etag: int):
        self.current_index = current_index
        self.current_node = current_node
        self.etag = etag


class NodeStatus:
    def __init__(
        self,
        name: str,
        connected: bool,
        error_details: str,
        last_send: datetime.datetime,
        last_reply: datetime.datetime,
        last_sent_message: str,
        last_matching_index: int,
    ):
        self.name = name
        self.connected = connected
        self.error_details = error_details
        self.last_send = last_send
        self.last_reply = last_reply
        self.last_sent_message = last_sent_message
        self.last_matching_index = last_matching_index

    def __str__(self):
        return (
            "NodeStatus{"
            f"name='{self.name}'"
            f", connected={self.connected}"
            f", errorDetails={self.error_details}"
            f", lastSend={self.last_send}"
            f", lastReply={self.last_reply}"
            f", lastSentMessage={self.last_sent_message}"
            "}"
        )


class RaftCommand:
    @abstractmethod
    def get_raft_unique_request_id(self) -> str:
        pass



from __future__ import annotations
from typing import Callable


class IncludesUtil:
    @staticmethod
    def include(document: dict, include: str, load_id: Callable[[str], None]):
        if not include or document is None:
            return

    @staticmethod
    def requires_quotes(include: str) -> (bool, str):
        for ch in include:
            if not ch.isalnum() and ch != "_" and ch != ".":
                return True, include.replace("'", "\\'")
        return False, include



from __future__ import annotations

import json
import logging
import socket
from enum import Enum
from typing import Optional, Tuple, Dict, Callable

from ravendb.documents.operations.replication.definitions import DetailedReplicationHubAccess


class TcpConnectionHeaderMessage:
    class OperationTypes(Enum):
        NONE = "None"
        DROP = "Drop"
        SUBSCRIPTION = "Subscription"
        REPLICATION = "REplication"
        CLUSTER = "Cluster"
        HEARTBEATS = "Heartbeats"
        PING = "Ping"
        TEST_CONNECTION = "TestConnection"

    def __init__(
        self,
        database_name: Optional[str] = None,
        source_node_tag: Optional[str] = None,
        operation: Optional[OperationTypes] = None,
        operation_version: Optional[int] = None,
        info: Optional[str] = None,
        authorize_info: Optional[AuthorizationInfo] = None,
        replication_hub_access: Optional[DetailedReplicationHubAccess] = None,
        server_id: Optional[str] = None,
    ):
        self.database_name = database_name
        self.source_node_tag = source_node_tag
        self.operation = operation
        self.operation_version = operation_version
        self.info = info
        self.authorize_info = authorize_info
        self.replication_hub_access = replication_hub_access
        self.server_id = server_id

    def to_json(self) -> Dict:
        return {
            "DatabaseName": self.database_name,
            "SourceNodeTag": self.source_node_tag,
            "Operation": self.operation,
            "OperationVersion": self.operation_version,
            "Info": self.info,
            "AuthorizeInfo": self.authorize_info.to_json(),
            "ReplicationHubAccess": self.replication_hub_access.to_json(),
            "ServerId": self.server_id,
        }

    NUMBER_OR_RETRIES_FOR_SENDING_TCP_HEADER = 2
    PING_BASE_LINE = -1
    NONE_BASE_LINE = -1
    DROP_BASE_LINE = -2
    HEARTBEATS_BASE_LINE = 20
    HEARTBEATS_41200 = 41_200
    HEARTBEATS_42000 = 42_000
    SUBSCRIPTION_BASE_LINE = 40
    SUBSCRIPTION_INCLUDES = 41_400
    SUBSCRIPTION_COUNTER_INCLUDES = 50_000
    SUBSCRIPTION_TIME_SERIES_INCLUDES = 51_000
    TEST_CONNECTION_BASE_LINE = 50

    HEARTBEATS_TCP_VERSION = HEARTBEATS_42000
    SUBSCRIPTION_TCP_VERSION = SUBSCRIPTION_TIME_SERIES_INCLUDES
    TEST_CONNECTION_TCP_VERSION = TEST_CONNECTION_BASE_LINE

    class SupportedFeatures:
        class PingFeatures:
            base_line = True

        class NoneFeatures:
            base_line = True

        class DropFeatures:
            base_line = True

        class SubscriptionFeatures:
            base_line = True

            def __init__(
                self,
                includes: Optional[bool] = None,
                counter_includes: Optional[bool] = None,
                time_series_includes: Optional[bool] = None,
            ):
                self.includes = includes
                self.counter_includes = counter_includes
                self.time_series_includes = time_series_includes

        class HeartbeatsFeatures:
            base_line = True

            def __init__(self, send_changes_only: Optional[bool] = None, include_server_info: Optional[bool] = None):
                self.send_changes_only = send_changes_only
                self.include_server_info = include_server_info

        class TestConnectionFeatures:
            base_line = True

        def __init__(
            self,
            version: int,
            ping: Optional[PingFeatures] = None,
            none: Optional[NoneFeatures] = None,
            drop: Optional[DropFeatures] = None,
            subscription: Optional[SubscriptionFeatures] = None,
            heartbeats: Optional[HeartbeatsFeatures] = None,
            test_connection: Optional[TestConnectionFeatures] = None,
        ):
            self.protocol_version = version
            self.ping = ping
            self.none = none
            self.drop = drop
            self.subscription = subscription
            self.heartbeats = heartbeats
            self.test_connection = test_connection

    operations_to_supported_protocol_versions = {
        OperationTypes.PING: [PING_BASE_LINE],
        OperationTypes.NONE: [NONE_BASE_LINE],
        OperationTypes.DROP: [DROP_BASE_LINE],
        OperationTypes.SUBSCRIPTION: [
            SUBSCRIPTION_TIME_SERIES_INCLUDES,
            SUBSCRIPTION_COUNTER_INCLUDES,
            SUBSCRIPTION_INCLUDES,
            SUBSCRIPTION_BASE_LINE,
        ],
        OperationTypes.HEARTBEATS: [HEARTBEATS_42000, HEARTBEATS_41200, HEARTBEATS_BASE_LINE],
        OperationTypes.TEST_CONNECTION: [TEST_CONNECTION_BASE_LINE],
    }

    supported_features_by_protocol = {
        OperationTypes.PING: {PING_BASE_LINE: SupportedFeatures(PING_BASE_LINE, SupportedFeatures.PingFeatures())},
        OperationTypes.NONE: {NONE_BASE_LINE: SupportedFeatures(NONE_BASE_LINE, none=SupportedFeatures.NoneFeatures())},
        OperationTypes.DROP: {DROP_BASE_LINE: SupportedFeatures(DROP_BASE_LINE, drop=SupportedFeatures.DropFeatures())},
        OperationTypes.SUBSCRIPTION: {
            SUBSCRIPTION_BASE_LINE: SupportedFeatures(
                SUBSCRIPTION_BASE_LINE, subscription=SupportedFeatures.SubscriptionFeatures()
            ),
            SUBSCRIPTION_INCLUDES: SupportedFeatures(
                SUBSCRIPTION_INCLUDES, subscription=SupportedFeatures.SubscriptionFeatures(True)
            ),
            SUBSCRIPTION_COUNTER_INCLUDES: SupportedFeatures(
                SUBSCRIPTION_COUNTER_INCLUDES, subscription=SupportedFeatures.SubscriptionFeatures(True, True)
            ),
            SUBSCRIPTION_TIME_SERIES_INCLUDES: SupportedFeatures(
                SUBSCRIPTION_TIME_SERIES_INCLUDES, subscription=SupportedFeatures.SubscriptionFeatures(True, True, True)
            ),
        },
        OperationTypes.HEARTBEATS: {
            HEARTBEATS_BASE_LINE: SupportedFeatures(
                HEARTBEATS_BASE_LINE, heartbeats=SupportedFeatures.HeartbeatsFeatures()
            ),
            HEARTBEATS_41200: SupportedFeatures(
                HEARTBEATS_41200, heartbeats=SupportedFeatures.HeartbeatsFeatures(True)
            ),
            HEARTBEATS_42000: SupportedFeatures(
                HEARTBEATS_42000, heartbeats=SupportedFeatures.HeartbeatsFeatures(True, True)
            ),
        },
        OperationTypes.TEST_CONNECTION: {
            TEST_CONNECTION_BASE_LINE: SupportedFeatures(
                TEST_CONNECTION_BASE_LINE, test_connection=SupportedFeatures.TestConnectionFeatures()
            )
        },
    }

    operations = [
        OperationTypes.CLUSTER,
        OperationTypes.DROP,
        OperationTypes.HEARTBEATS,
        OperationTypes.NONE,
        OperationTypes.PING,
        OperationTypes.REPLICATION,
        OperationTypes.SUBSCRIPTION,
        OperationTypes.TEST_CONNECTION,
    ]

    class SupportedStatus(Enum):
        OUT_OF_RANGE = "OutOfRange"
        NOT_SUPPORTED = "NotSupported"
        SUPPORTED = "Supported"

    @staticmethod
    def operation_version_supported(operation_type: OperationTypes, version: int) -> Tuple[SupportedStatus, int]:
        current = -1
        supported_protocols = TcpConnectionHeaderMessage.operations_to_supported_protocol_versions.get(operation_type)
        if supported_protocols is None:
            raise RuntimeError(
                f"This is a bug. Probably you forgot to add '{operation_type}' operation to the operations_to_supported_protocol_versions dict"
            )

        for ver in supported_protocols:
            current = ver
            if ver == version:
                return TcpConnectionHeaderMessage.SupportedStatus.SUPPORTED, current

            if ver < version:
                return TcpConnectionHeaderMessage.SupportedStatus.NOT_SUPPORTED, current

        return TcpConnectionHeaderMessage.SupportedStatus.OUT_OF_RANGE, current

    @staticmethod
    def get_operation_tcp_version(operation_type: OperationTypes, index: int) -> int:
        # we don't check the if the index go out of range, since this is expected and means that we don't have
        if operation_type in [
            TcpConnectionHeaderMessage.OperationTypes.PING,
            TcpConnectionHeaderMessage.OperationTypes.NONE,
        ]:
            return -1
        if operation_type == TcpConnectionHeaderMessage.OperationTypes.DROP:
            return -2
        if operation_type in [
            TcpConnectionHeaderMessage.OperationTypes.SUBSCRIPTION,
            TcpConnectionHeaderMessage.OperationTypes.REPLICATION,
            TcpConnectionHeaderMessage.OperationTypes.CLUSTER,
            TcpConnectionHeaderMessage.OperationTypes.HEARTBEATS,
            TcpConnectionHeaderMessage.OperationTypes.TEST_CONNECTION,
        ]:
            return TcpConnectionHeaderMessage.operations_to_supported_protocol_versions.get(operation_type)[index]
        raise ValueError("operation_type")

    @staticmethod
    def get_supported_features_for(op_type: OperationTypes, protocol_version: int) -> SupportedFeatures:
        features = TcpConnectionHeaderMessage.supported_features_by_protocol.get(op_type).get(protocol_version)
        if features is None:
            raise ValueError(f"{op_type} in protocol {protocol_version} was not found in the features set")
        return features

    class AuthorizationInfo:
        class AuthorizeMethod(Enum):
            SERVER = "Server"
            PULL_REPLICATION = "PullReplication"
            PUSH_REPLICATION = "PushReplication"

        def __init__(self, authorize_as: Optional[AuthorizeMethod] = None, authorization_for: Optional[str] = None):
            self.authorize_as = authorize_as
            self.authorization_for = authorization_for

        def to_json(self) -> Dict:
            return {"AuthorizeAs": self.authorize_as.value, "AuthorizationFor": self.authorization_for}


class TcpNegotiateParameters:
    def __init__(
        self,
        operation: Optional[TcpConnectionHeaderMessage.OperationTypes] = None,
        authorize_info: Optional[TcpConnectionHeaderMessage.AuthorizationInfo] = None,
        version: Optional[int] = None,
        database: Optional[str] = None,
        source_node_tag: Optional[str] = None,
        destination_node_tag: Optional[str] = None,
        destination_url: Optional[str] = None,
        read_response_and_get_version_callback: Optional[Callable[[str, socket.socket], int]] = None,
        destination_server_id: Optional[str] = None,
    ):
        self.operation = operation
        self.authorize_info = authorize_info
        self.version = version
        self.database = database
        self.source_node_tag = source_node_tag
        self.destination_node_tag = destination_node_tag
        self.destination_url = destination_url
        self.read_response_and_get_version_callback = read_response_and_get_version_callback
        self.destination_server_id = destination_server_id


class TcpNegotiation:
    logger = logging.getLogger("TcpNegotiation")
    OUT_OF_RANGE_STATUS = -1
    DROP_STATUS = -2

    @classmethod
    def negotiate_protocol_version(
        cls, sock: socket.socket, parameters: TcpNegotiateParameters
    ) -> TcpConnectionHeaderMessage.SupportedFeatures:
        cls.logger.info(
            f"Start of negotiation for {parameters.operation} operation with {parameters.destination_node_tag or parameters.destination_url}"
        )
        current = parameters.version
        while True:
            cls._send_tcp_version_info(sock, parameters, current)
            version = parameters.read_response_and_get_version_callback(parameters.destination_url, sock)

            cls.logger.info(
                f"Read response from {parameters.source_node_tag or parameters.destination_url} "
                f"for {parameters.operation}, received version is '{version}'"
            )

            if version == current:
                break

            if version == cls.DROP_STATUS:
                return TcpConnectionHeaderMessage.get_supported_features_for(
                    TcpConnectionHeaderMessage.OperationTypes.DROP, TcpConnectionHeaderMessage.DROP_BASE_LINE
                )

            status, current = TcpConnectionHeaderMessage.operation_version_supported(parameters.operation, version)
            if status == TcpConnectionHeaderMessage.SupportedStatus.OUT_OF_RANGE:
                cls._send_tcp_version_info(sock, parameters, cls.OUT_OF_RANGE_STATUS)
                raise ValueError(
                    f"The {parameters.operation} version {parameters.version} is out of range, out lowest version is {current}"
                )

            cls.logger.info(
                f"The version {version} is {status}, will try to agree on '{current}' for {parameters.operation} with {parameters.destination_node_tag or parameters.destination_url}"
            )

        cls.logger.info(
            f"{parameters.destination_node_tag or parameters.destination_url} agreed on version {current} for {parameters.operation}"
        )

        return TcpConnectionHeaderMessage.get_supported_features_for(parameters.operation, current)

    @classmethod
    def _send_tcp_version_info(
        cls, sock: socket.socket, parameters: TcpNegotiateParameters, current_version: int
    ) -> None:
        cls.logger.info(f"Send negotiation for {parameters.operation} in version {current_version}")
        json_dict = {
            "DatabaseName": parameters.database,
            "Operation": parameters.operation.value,
            "SourceNodeTag": parameters.source_node_tag,
            "OperationVersion": current_version,
            "AuthorizeInfo": parameters.authorize_info.to_json() if parameters.authorize_info is not None else None,
        }

        sock.send(json.dumps(json_dict).encode("utf-8"))


class TcpConnectionStatus(Enum):
    OK = "Ok"
    AUTHORIZATION_FAILED = "AuthorizationFailed"
    TCP_VERSION_MISMATCH = "TcpVersionMismatch"
    INVALID_NETWORK_TOPOLOGY = "InvalidNetworkTopology"


class TcpConnectionHeaderResponse:
    def __init__(self, status: TcpConnectionStatus, message: str, version: int):
        self.status = status
        self.message = message
        self.version = version

    @classmethod
    def from_json(cls, json_dict: Dict) -> TcpConnectionHeaderResponse:
        return cls(
            TcpConnectionStatus(json_dict.get("Status", None)),
            json_dict.get("Message", None),
            json_dict.get("Version", None),
        )



from decimal import InvalidOperation
from typing import Any, Dict, Optional

from ijson.common import integer_or_decimal, IncompleteJSONError
from ijson.backends.python import UnexpectedSymbol
from _elementtree import ParseError
from socket import timeout
import re

# BUFSIZE isn't included in newer ijson versions so we define it ourselves:
# See: https://github.com/isagalaev/ijson/blob/c594cdd3c94c8b4a018577966b3ad22bb44c2620/ijson/backends/python.py#L14
BUFSIZE = 16 * 1024
LEXEME_RE = re.compile(b"[a-z0-9eE\.\+-]+|\S")
BYTE_ARRAY_CHARACTERS = bytearray(b',}:{"')
IS_WEBSOCKET = False


# The code imported from ijson to be able to receive json from socket
class IncrementalJsonParser:
    def __init__(self, socket, is_websocket=False):
        global IS_WEBSOCKET
        IS_WEBSOCKET = is_websocket
        self.lexer = self.lexer(socket)

    @staticmethod
    def lexer(socket, buf_size=BUFSIZE):
        data = socket.recv() if IS_WEBSOCKET else socket.recv(buf_size)
        if not data:
            return

        buf = data[1:-1].encode("utf-8") if IS_WEBSOCKET else data
        pos = 0
        discarded = 0
        while True:
            match = LEXEME_RE.search(buf, pos)
            if pos < len(buf) and match:
                lexeme = match.group()
                if lexeme == b'"':
                    pos = match.start()
                    start = pos + 1
                    while True:
                        try:
                            end = buf.index(b'"', start)
                            escpos = end - 1
                            while buf[escpos] == "\\":
                                escpos -= 1
                            if (end - escpos) % 2 == 0:
                                start = end + 1
                            else:
                                break
                        except ValueError:
                            data = socket.recv().encode("utf-8") if IS_WEBSOCKET else socket.recv(buf_size)
                            if not data:
                                raise IncompleteJSONError("Incomplete string lexeme")

                            buf += data[1:-1] if IS_WEBSOCKET else data

                    yield discarded + pos, buf[pos : end + 1].decode("utf-8")
                    pos = end + 1
                else:
                    while match.end() == len(buf) and buf[pos] not in BYTE_ARRAY_CHARACTERS:
                        try:
                            data = socket.recv().encode("utf-8") if IS_WEBSOCKET else socket.recv(buf_size)
                            if not data:
                                break
                            buf += data[1:-1].encode("utf-8") if IS_WEBSOCKET else data
                            match = LEXEME_RE.search(buf, pos)
                            lexeme = match.group()
                        except timeout:
                            break

                    yield match.start(), lexeme.decode("utf-8")
                    pos = match.end()
            else:
                data = socket.recv().encode("utf-8") if IS_WEBSOCKET else socket.recv(buf_size)
                if not data:
                    break
                discarded += len(buf)
                buf = data[1:-1] if IS_WEBSOCKET else data
                pos = 0

    def create_array(self, gen):
        arr = []

        for token, val in gen:
            if token == "end_array":
                return arr
            arr.append(self.get_value_from_token(gen, token, val))

        raise ParseError("End array expected, but the generator ended before we got it")

    def get_value(self, gen):
        (token, val) = next(gen)

        return self.get_value_from_token(gen, token, val)

    def get_value_from_token(self, gen, token, val):
        if token == "start_array":
            return self.create_array(gen)

        if token == "start_map":
            return self.create_object(gen)

        return val

    def create_object(self, gen):
        obj = {}

        for token, val in gen:
            if token == "end_map":
                return obj
            if token == "map_key":
                obj[val] = self.get_value(gen)

        raise ParseError("End object expected, but the generator ended before we got it")

    def next_object(self) -> Optional[Dict[str, Any]]:
        try:
            (_, text) = next(self.lexer)
            if IS_WEBSOCKET and text == ",":
                (_, text) = next(self.lexer)
        except StopIteration:
            return None

        if text != "{":
            raise ParseError("Expected start object, got: " + text)

        gen = IncrementalJsonParser.parse_object(self.lexer)
        (token, val) = next(gen)
        assert token == "start_map"

        return self.create_object(gen)

    @staticmethod
    def parse_value(lexer, symbol=None, pos=0):
        try:
            if symbol is None:
                pos, symbol = next(lexer)
            if symbol == "null":
                yield ("null", None)
            elif symbol == "true":
                yield ("boolean", True)
            elif symbol == "false":
                yield ("boolean", False)
            elif symbol == "[":
                for event in IncrementalJsonParser.parse_array(lexer):
                    yield event
            elif symbol == "{":
                for event in IncrementalJsonParser.parse_object(lexer):
                    yield event
            elif symbol[0] == '"':
                yield ("string", IncrementalJsonParser.unescape(symbol[1:-1]))
            else:
                # if we got a partial token for false / null / true we need to read from the network again
                while symbol[0] in ("t", "n") and len(symbol) < 4 or symbol[0] == "f" and len(symbol) < 5:
                    _, nextpart = next(lexer)
                    if symbol == "null":
                        yield ("null", None)
                    elif symbol == "true":
                        yield ("boolean", True)
                    elif symbol == "false":
                        yield ("boolean", False)
                    return

                try:
                    yield ("number", integer_or_decimal(symbol))
                except InvalidOperation:
                    raise UnexpectedSymbol(symbol, pos)
        except StopIteration:
            raise IncompleteJSONError("Incomplete JSON data")

    @staticmethod
    def parse_array(lexer):
        yield ("start_array", None)
        try:
            pos, symbol = next(lexer)
            if symbol != "]":
                while True:
                    for event in IncrementalJsonParser.parse_value(lexer, symbol, pos):
                        yield event
                    pos, symbol = next(lexer)
                    if symbol == "]":
                        break
                    if symbol != ",":
                        raise UnexpectedSymbol(symbol, pos)
                    pos, symbol = next(lexer)
            yield ("end_array", None)
        except StopIteration:
            raise IncompleteJSONError("Incomplete JSON data")

    @staticmethod
    def parse_object(lexer):
        yield ("start_map", None)
        try:
            pos, symbol = next(lexer)
            if symbol != "}":
                while True:
                    if symbol[0] != '"':
                        raise UnexpectedSymbol(symbol, pos)
                    yield ("map_key", IncrementalJsonParser.unescape(symbol[1:-1]))
                    pos, symbol = next(lexer)
                    if symbol != ":":
                        raise UnexpectedSymbol(symbol, pos)
                    for event in IncrementalJsonParser.parse_value(lexer, None, pos):
                        yield event
                    pos, symbol = next(lexer)
                    if symbol == "}":
                        break
                    if symbol != ",":
                        raise UnexpectedSymbol(symbol, pos)
                    pos, symbol = next(lexer)
            yield ("end_map", None)
        except StopIteration:
            raise IncompleteJSONError("Incomplete JSON data")

    @staticmethod
    def unescape(s):
        start = 0
        result = ""
        while start < len(s):
            pos = s.find("\\", start)
            if pos == -1:
                if start == 0:
                    return s
                result += s[start:]
                break
            result += s[start:pos]
            pos += 1
            esc = s[pos]
            if esc == "u":
                result += chr(int(s[pos + 1 : pos + 5], 16))
                pos += 4
            elif esc == "b":
                result += "\b"
            elif esc == "f":
                result += "\f"
            elif esc == "n":
                result += "\n"
            elif esc == "r":
                result += "\r"
            elif esc == "t":
                result += "\t"
            else:
                result += esc
            start = pos + 1
        return result



from __future__ import annotations

import http
import json
from typing import Optional, TYPE_CHECKING, List

import requests

from ravendb.primitives import constants
from ravendb.data.operation import AttachmentType
from ravendb.documents.operations.definitions import IOperation, VoidOperation
from ravendb.http.http_cache import HttpCache
from ravendb.http.misc import ResponseDisposeHandling
from ravendb.http.raven_command import RavenCommand, RavenCommandResponseType, VoidRavenCommand
from ravendb.http.server_node import ServerNode
from ravendb.tools.utils import Utils

if TYPE_CHECKING:
    from ravendb.documents import DocumentStore
    from ravendb.documents.conventions import DocumentConventions


class AttachmentName:
    def __init__(self, name: str, hash: str, content_type: str, size: int):
        self.name = name
        self.hash = hash
        self.content_type = content_type
        self.size = size

    @classmethod
    def from_json(cls, json_dict: dict) -> AttachmentName:
        return cls(json_dict["Name"], json_dict["Hash"], json_dict["ContentType"], json_dict["Size"])


class AttachmentDetails(AttachmentName):
    def __init__(
        self, name: str, hash: str, content_type: str, size: int, change_vector: str = None, document_id: str = None
    ):
        super().__init__(name, hash, content_type, size)
        self.change_vector = change_vector
        self.document_id = document_id

    @classmethod
    def from_json(cls, json_dict: dict) -> AttachmentDetails:
        return cls(
            json_dict["Name"],
            json_dict["Hash"],
            json_dict["ContentType"],
            json_dict["Size"],
            json_dict["ChangeVector"],
            json_dict["DocumentId"],
        )


class CloseableAttachmentResult:
    def __init__(self, response: requests.Response, details: AttachmentDetails):
        self.__details = details
        self.__response = response

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @property
    def data(self):
        return self.__response.content

    @property
    def details(self):
        return self.__details

    def close(self):
        self.__response.close()


class AttachmentRequest:
    def __init__(self, document_id: str, name: str):
        if document_id.isspace():
            raise ValueError("Document_id cannot be None or whitespace")
        if name.isspace():
            raise ValueError("Name cannot be None or whitespace")

        self.document_id = document_id
        self.name = name


class PutAttachmentOperation(IOperation[AttachmentDetails]):
    def __init__(
        self,
        document_id: str,
        name: str,
        stream: bytes,
        content_type: Optional[str] = None,
        change_vector: Optional[str] = None,
    ):
        super().__init__()
        self.__document_id = document_id
        self.__name = name
        self.__stream = stream
        self.__content_type = content_type
        self.__change_vector = change_vector

    def get_command(self, store, conventions, cache=None):
        return self.__PutAttachmentCommand(
            self.__document_id,
            self.__name,
            self.__stream,
            self.__content_type,
            self.__change_vector,
        )

    class __PutAttachmentCommand(RavenCommand[AttachmentDetails]):
        def __init__(self, document_id: str, name: str, stream: bytes, content_type: str, change_vector: str):
            super().__init__(AttachmentDetails)

            if not document_id:
                raise ValueError("document_id cannot be None or blank")

            if not name:
                raise ValueError("name cannot be None or blank")

            self.__document_id = document_id
            self.__name = name
            self.__stream = stream
            self.__content_type = content_type
            self.__change_vector = change_vector

        def create_request(self, node: ServerNode) -> requests.Request:
            url = (
                f"{node.url}/databases/{node.database}/attachments?id={self.__document_id}"  # todo : escape
                f"&name={Utils.escape(self.__name, True,False)}"
            )

            if not self.__content_type.isspace():
                url += f"&contentType={Utils.escape(self.__content_type, True, False)}"

            request = requests.Request("PUT", url)
            if isinstance(self.__stream, (bytes, bytearray)):
                request.files = {self.__name: (self.__name, self.__stream, self.__content_type)}
            else:
                request.data = self.__stream
            self._add_change_vector_if_not_none(self.__change_vector, request)
            return request

        def set_response(self, response: str, from_cache: bool) -> None:
            self.result = AttachmentDetails.from_json(json.loads(response))

        def is_read_request(self) -> bool:
            return False


class GetAttachmentOperation(IOperation):
    def __init__(self, document_id: str, name: str, attachment_type: AttachmentType, change_vector: Optional[str]):
        if document_id is None:
            raise ValueError("Invalid document_id")
        if name is None:
            raise ValueError("Invalid name")

        if attachment_type != AttachmentType.document and change_vector is None:
            raise ValueError(
                "change_vector",
                "Change Vector cannot be null for attachment type {0}".format(attachment_type),
            )

        super(GetAttachmentOperation, self).__init__()
        self.__document_id = document_id
        self.__name = name
        self.__attachment_type = attachment_type
        self.__change_vector = change_vector

    def get_command(self, store, conventions, cache=None) -> RavenCommand[dict]:
        return self.__GetAttachmentCommand(
            self.__document_id, self.__name, self.__attachment_type, self.__change_vector
        )

    class __GetAttachmentCommand(RavenCommand[CloseableAttachmentResult]):
        def __init__(self, document_id: str, name: str, attachment_type: AttachmentType, change_vector: str):
            super().__init__(CloseableAttachmentResult)
            self.__document_id = document_id
            self.__name = name
            self.__attachment_type = attachment_type
            self.__change_vector = change_vector

        def create_request(self, server_node: ServerNode) -> requests.Request:
            method = "GET"
            url = (
                f"{server_node.url}/databases/{server_node.database}/attachments"
                f"?id={Utils.quote_key(self.__document_id)}&name={Utils.quote_key(self.__name)}"
            )
            data = None
            if self.__attachment_type != AttachmentType.document:
                method = "POST"
                data = {
                    "Type": str(self.__attachment_type),
                    "ChangeVector": self.__change_vector,
                }
            return requests.Request(method, url, data=data)

        def is_read_request(self) -> bool:
            return True

        def process_response(self, cache: HttpCache, response: requests.Response, url) -> http.ResponseDisposeHandling:
            content_type = response.headers.get("Content-Type")
            change_vector = response.headers.get(constants.Headers.ETAG)
            hash = response.headers.get("Attachment-Hash")
            size = response.headers.get("Attachment-Size", 0)
            attachment_details = AttachmentDetails(
                self.__name, hash, content_type, size, change_vector, self.__document_id
            )
            self.result = CloseableAttachmentResult(response, attachment_details)
            return ResponseDisposeHandling.MANUALLY


class GetAttachmentsOperation(IOperation[CloseableAttachmentResult]):  # unfinished
    def __init__(self, attachments: List[AttachmentRequest], attachment_type: AttachmentType):
        self.__attachment_type = attachment_type
        self.__attachments = attachments

    def get_command(
        self, store: DocumentStore, conventions: DocumentConventions, cache: HttpCache
    ) -> RavenCommand[CloseableAttachmentResult]:
        return self.__GetAttachmentsCommand(self.__attachments, self.__attachment_type)

    class __GetAttachmentsCommand(RavenCommand[CloseableAttachmentResult]):
        def __init__(self, attachments: List[AttachmentRequest], attachment_type: AttachmentType):
            super().__init__(CloseableAttachmentResult)
            self.__attachment_type = attachment_type
            self.__attachments = attachments
            self._response_type = RavenCommandResponseType.EMPTY
            self.__attachments_metadata = []

        def create_request(self, node: ServerNode) -> requests.Request:
            return requests.Request(
                "POST",
                f"{node.url}/databases/{node.database}/attachments/bulk",
                data={
                    "AttachmentType": str(self.__attachment_type),
                    "Attachments": [
                        {"DocumentId": attachment.document_id, "Name": attachment.name}
                        for attachment in self.__attachments
                    ],
                },
            )

        def is_read_request(self) -> bool:
            return True

        def process_response(self, cache: HttpCache, response: requests.Response, url) -> ResponseDisposeHandling:
            raise NotImplementedError("Fetching multiple attachments is yet to be implemented")
            # todo: we can't decode it - maybe split bytes (part 1 is json, second part is binary attachments data) -


class DeleteAttachmentOperation(VoidOperation):
    def __init__(self, document_id: str, name: str, change_vector: Optional[str] = None):
        self._document_id = document_id
        self._name = name
        self._change_vector = change_vector

    def get_command(self, store: "DocumentStore", conventions: "DocumentConventions", cache=None) -> VoidRavenCommand:
        return self.__DeleteAttachmentCommand(self._document_id, self._name, self._change_vector)

    class __DeleteAttachmentCommand(VoidRavenCommand):
        def __init__(self, document_id: str, name: str, change_vector: Optional[str] = None):
            super().__init__()
            if not document_id:
                raise ValueError("Invalid document_id")
            if not name:
                raise ValueError("Invalid name")

            self.__document_id = document_id
            self.__name = name
            self.__change_vector = change_vector

        def create_request(self, server_node):
            request = requests.Request(
                "DELETE",
                f"{server_node.url}/databases/{server_node.database}/attachments?id={Utils.quote_key(self.__document_id)}&name={Utils.quote_key(self.__name)}",
            )
            self._add_change_vector_if_not_none(self.__change_vector, request)
            return request



from __future__ import annotations
import threading
from typing import Generic, Callable, Any, Union, TypeVar

from ravendb.documents.commands.batches import CommandType

_T = TypeVar("_T")


class IdTypeAndName:
    def __init__(self, key: str, command_type: CommandType, name: str):
        self.key = key
        self.command_type = command_type
        self.name = name

    def __eq__(self, other):
        if other is None or type(self) != type(other):
            return False
        other: IdTypeAndName
        if self.key != other.key if self.key else other.key:
            return False
        if self.command_type != other.command_type:
            return False
        return self.name == other.name if self.name is not None else other.name is None

    def __hash__(self):
        return hash((self.key, self.command_type, self.name))

    @classmethod
    def create(cls, key: str, command_type: CommandType, name: Union[None, str]) -> IdTypeAndName:
        return cls(key, command_type, name)



from __future__ import annotations
from typing import Generic, TypeVar, Optional

_T = TypeVar("_T")


class ConditionalLoadResult(Generic[_T]):
    def __init__(self):
        self._entity = None
        self._change_vector = None

    @property
    def entity(self) -> Optional[_T]:
        return self._entity

    @property
    def change_vector(self) -> Optional[str]:
        return self._change_vector

    @classmethod
    def create(cls, entity: Optional[_T], change_vector: Optional[str]) -> ConditionalLoadResult[Optional[_T]]:
        result = cls()
        result._entity = entity
        result._change_vector = change_vector
        return result



import json
from typing import Dict, Type, TypeVar, Optional

from ravendb.primitives import constants
from ravendb.documents.conventions import DocumentConventions
from ravendb.documents.operations.compare_exchange.compare_exchange import CompareExchangeValue
from ravendb.json.metadata_as_dictionary import MetadataAsDictionary
from ravendb.tools.utils import CaseInsensitiveDict, Utils

_T = TypeVar("_T")


class CompareExchangeValueResultParser:
    @staticmethod
    def get_values(
        object_type: Type[_T], response: str, materialize_metadata: bool, conventions: DocumentConventions
    ) -> Dict[str, CompareExchangeValue[_T]]:
        results = CaseInsensitiveDict()
        if not response:
            return results
        items = json.loads(response)["Results"]
        if not items:
            raise ValueError("Response is invalid. Results is missing.")

        for item in items:
            if not item:
                raise ValueError("Response is invalid. Item is None")
            value = CompareExchangeValueResultParser.get_single_value(
                object_type, item, materialize_metadata, conventions
            )
            results[value.key] = value

        return results

    # todo: check if we can't set object type as optional
    @staticmethod
    def get_value(
        object_type: Type[_T], response: str, materialize_metadata: bool, conventions: DocumentConventions
    ) -> Optional[CompareExchangeValue[_T]]:
        if not response or response == "null":
            return None

        values = CompareExchangeValueResultParser.get_values(object_type, response, materialize_metadata, conventions)
        if not values:
            return None
        return values.values().__iter__().__next__()

    @staticmethod
    def get_single_value(
        object_type: Type[_T], item: dict, materialize_metadata: bool, conventions: DocumentConventions
    ) -> Optional[CompareExchangeValue[_T]]:
        if item is None:
            return None
        if "Key" not in item:
            raise KeyError("Response is invalid. Key is missing.")
        if "Index" not in item:
            raise IndexError("Response is invalid. Index is missing.")
        if "Value" not in item:
            raise ValueError("Response is invalid. Value is missing.")

        key: str = item.get("Key")
        index: int = item.get("Index")
        raw: dict = item.get("Value")
        if not raw:
            return CompareExchangeValue(key, index, None)
        metadata = None
        bjro = raw.get(constants.Documents.Metadata.KEY)
        if bjro:
            metadata = (
                MetadataAsDictionary(bjro)
                if not materialize_metadata
                else MetadataAsDictionary.materialize_from_json(bjro)
            )
        if object_type in Utils.primitives_and_collections:
            value = None
            if raw:
                raw_value = raw.get("Object")
                value = (
                    Utils.convert_json_dict_to_object(raw_value, object_type)
                    if not isinstance(raw_value, Utils.primitives_and_collections)
                    else raw_value
                )

            return CompareExchangeValue(key, index, value, metadata)
        elif object_type == dict:
            if not raw or constants.CompareExchange.OBJECT_FIELD_NAME not in raw:
                return CompareExchangeValue(key, index, None, metadata)

            raw_value = raw.get(constants.CompareExchange.OBJECT_FIELD_NAME)
            if not raw:
                return CompareExchangeValue(key, index, None, metadata)
            else:
                return CompareExchangeValue(key, index, raw_value, metadata)
        else:
            obj = raw.get(constants.CompareExchange.OBJECT_FIELD_NAME)
            if not obj:
                return CompareExchangeValue(key, index, None, metadata)
            else:
                converted = (
                    Utils.convert_json_dict_to_object(obj, object_type)
                    if not isinstance(obj, Utils.primitives_and_collections)
                    else obj
                )
                return CompareExchangeValue(key, index, converted, metadata)



import inspect
from copy import deepcopy
from typing import Optional, TYPE_CHECKING, Union, Type, TypeVar, Dict, Any, Tuple

from ravendb.primitives import constants
from ravendb.documents.session.document_info import DocumentInfo
from ravendb.documents.session.event_args import (
    BeforeConversionToDocumentEventArgs,
    AfterConversionToDocumentEventArgs,
    BeforeConversionToEntityEventArgs,
    AfterConversionToEntityEventArgs,
)
from ravendb.exceptions import exceptions
from ravendb.exceptions.exceptions import InvalidOperationException
from ravendb.tools.utils import Utils, DynamicStructure
from ravendb.documents.conventions import DocumentConventions

if TYPE_CHECKING:
    from ravendb.documents.session.document_session_operations.in_memory_document_session_operations import (
        InMemoryDocumentSessionOperations,
    )

_T = TypeVar("_T")


class EntityToJson:
    def __init__(self, session: "InMemoryDocumentSessionOperations"):
        self._session = session
        self._missing_dictionary = dict()

    @property
    def missing_dictionary(self):
        return self._missing_dictionary

    # Converting objects to JSON Dicts
    # ================================
    def convert_entity_to_json(self, entity: object, document_info: DocumentInfo) -> dict:
        should_invoke_events = document_info is not None
        if should_invoke_events:
            self._session.before_conversion_to_document_invoke(
                BeforeConversionToDocumentEventArgs(document_info.key, entity, self._session)
            )

        document = EntityToJson._convert_entity_to_json_internal(self, entity, document_info, True)

        if should_invoke_events:
            self._session.after_conversion_to_document_invoke(
                AfterConversionToDocumentEventArgs(self._session, document_info.key, entity, document)
            )
        return document

    def _convert_entity_to_json_internal(
        self, entity: object, document_info: DocumentInfo, remove_identity_property: bool = False
    ) -> dict:
        json_node = Utils.entity_to_dict(entity, self._session.conventions.json_default_method)
        EntityToJsonUtils.write_metadata(json_node, document_info)

        if remove_identity_property:
            EntityToJsonUtils.try_remove_identity_property_from_json_dict(
                json_node, entity.__class__, self._session.conventions
            )

        return json_node

    # Converting JSON Dicts to objects
    # ================================
    def convert_to_entity(self, entity_type: Type[_T], key: str, document: dict, track_entity: bool) -> _T:
        conventions = self._session.conventions
        return EntityToJsonStatic.convert_to_entity(document, entity_type, conventions, self._session, key)

    # Miscellaneous
    # =============
    def populate_entity(self, entity: object, key: str, document: Dict[str, Any]) -> None:
        if key is None:
            raise ValueError("Key cannot be None")

        EntityToJsonStatic.populate_entity(entity, document)
        self._session.generate_entity_id_on_the_client.try_set_identity(entity, key)

    def remove_from_missing(self, entity):
        try:
            del self.missing_dictionary[entity]
        except KeyError:
            pass

    def clear(self):
        self.missing_dictionary.clear()


class EntityToJsonStatic:
    @staticmethod
    def populate_entity(entity: object, document: Dict[str, Any]) -> None:
        if entity is None:
            raise ValueError("Entity cannot be None")
        if document is None:
            raise ValueError("Document cannot be None")
        entity.__dict__.update(document)

    @staticmethod
    def convert_entity_to_json(
        entity: object,
        conventions: "DocumentConventions",
        document_info: Optional[DocumentInfo],
        remove_identity_property: Optional[bool] = False,
    ) -> Dict[str, Any]:
        json_dict = Utils.entity_to_dict(entity, conventions.json_default_method)
        EntityToJsonUtils.write_metadata(json_dict, document_info)

        if remove_identity_property:
            EntityToJsonUtils.try_remove_identity_property_from_json_dict(json_dict, entity.__class__, conventions)

        return json_dict

    @staticmethod
    def convert_to_entity_by_key(
        entity_class: Type[_T], key: str, document: Dict, conventions: DocumentConventions
    ) -> _T:
        if entity_class is None:
            return document
        try:
            default_value = Utils.get_default_value(entity_class)
            entity = default_value

            document_type = conventions.get_python_class(key, document)
            if document_type is not None:
                clazz = Utils.import_class(document_type)
                if clazz is not None and issubclass(clazz, entity_class):
                    entity = EntityToJsonStatic.convert_to_entity(document, clazz, conventions)

            if entity is None:
                entity = EntityToJsonStatic.convert_to_entity(document, entity_class, conventions)

            return entity
        except Exception as e:
            raise RuntimeError(f"Could not convert document {key} to entity of type {entity_class}", e)

    @staticmethod
    def convert_to_entity(
        document: Dict[str, Any],
        object_type: Type[_T],
        conventions: "DocumentConventions",
        session: Optional["InMemoryDocumentSessionOperations"] = None,
        key: str = None,
    ) -> _T:
        metadata = document.get("@metadata")
        document_deepcopy = deepcopy(document)

        object_type, is_projection, should_update_metadata_python_type = EntityToJsonUtils.determine_object_type(
            document, conventions, object_type, metadata
        )

        if object_type is dict:
            EntityToJsonUtils.invoke_after_conversion_to_entity_event(session, key, object_type, document_deepcopy)
            return document_deepcopy

        if object_type is DynamicStructure:
            dyn = DynamicStructure(**document_deepcopy)
            EntityToJsonUtils.invoke_after_conversion_to_entity_event(session, key, object_type, document_deepcopy)
            return dyn

        if should_update_metadata_python_type:
            EntityToJsonUtils.set_python_type_in_metadata(metadata, object_type)

        # Fire before conversion to entity events
        if session:
            session.before_conversion_to_entity_invoke(
                BeforeConversionToEntityEventArgs(session, key, object_type, document_deepcopy)
            )

        # Conversion to entity

        if "from_json" in object_type.__dict__ and inspect.ismethod(object_type.from_json):
            # By custom defined 'from_json' serializer class method
            entity = object_type.from_json(document_deepcopy)
        elif is_projection:
            entity = DynamicStructure(**document_deepcopy)
            entity.__class__ = object_type
            try:
                entity = Utils.initialize_object(document_deepcopy, object_type)
            except TypeError as e:
                raise InvalidOperationException("Probably projection error", e)
        else:
            entity = Utils.convert_json_dict_to_object(document_deepcopy, object_type)

        EntityToJsonUtils.invoke_after_conversion_to_entity_event(session, key, object_type, document_deepcopy)

        # Try to set identity property
        identity_property_name = conventions.get_identity_property_name(object_type)
        if identity_property_name in entity.__dict__:
            entity.__dict__[identity_property_name] = metadata.get("@id", None)

        return entity


class EntityToJsonUtils:
    @staticmethod
    def invoke_after_conversion_to_entity_event(
        session: Optional["InMemoryDocumentSessionOperations"],
        key: str,
        object_type: Optional[_T],
        document_deepcopy: dict,
    ):
        if session:
            session.after_conversion_to_entity_invoke(
                AfterConversionToEntityEventArgs(session, key, object_type, document_deepcopy)
            )

    @staticmethod
    def try_remove_identity_property_from_json_dict(
        document: Dict[str, Any], entity_type: Type[Any], conventions: DocumentConventions
    ) -> bool:
        identity_property_name = conventions.get_identity_property_name(entity_type)
        if identity_property_name is None:
            return False

        try:
            del document[identity_property_name]
            return True
        except KeyError:
            return False

    @staticmethod
    def write_metadata(json_node: dict, document_info: DocumentInfo):
        if document_info is None:
            return
        set_metadata = False
        metadata_node = {}

        if document_info.metadata and len(document_info.metadata) > 0:
            set_metadata = True
            for name, value in document_info.metadata.items():
                metadata_node.update({name: deepcopy(value)})
        elif document_info.metadata_instance:
            set_metadata = True
            for key, value in document_info.metadata_instance.items():
                metadata_node.update({key: value})

        if document_info.collection:
            set_metadata = True
            metadata_node.update({constants.Documents.Metadata.COLLECTION: document_info.collection})

        if set_metadata:
            json_node.update({constants.Documents.Metadata.KEY: metadata_node})

    @staticmethod
    def determine_object_type(
        document: Dict[str, Any],
        conventions: DocumentConventions,
        object_type_from_user: Optional[Type[Any]] = None,
        metadata: Dict[str, Any] = None,
    ) -> Tuple[
        Type[Union[Any, DynamicStructure, Dict[str, Any]]], bool, bool
    ]:  # -> object_type, is_projection, should_update_metadata_python_type
        # Try to extract the object type from the metadata
        type_name_from_metadata = conventions.try_get_type_from_metadata(metadata)

        # Check if user needs dictionary or if we can return dictionary
        if object_type_from_user is dict or (
            object_type_from_user is None and type_name_from_metadata == "builtins.dict"
        ):
            return dict, False, False

        # No Python type in metadata
        if type_name_from_metadata is None:
            if object_type_from_user is not None:
                # Try using passed object type
                return object_type_from_user, False, True
            else:
                # No type defined, but the user didn't explicitly say that they need a dict - return DynamicStructure
                return DynamicStructure, False, False

        # Python object type is in the metadata
        object_type_from_metadata = Utils.import_class(type_name_from_metadata)
        if object_type_from_metadata is None:
            # Unsuccessful import means the document has been probably stored within other Python project
            # or the original object class has been removed - essentially we have only object_type to rely on
            if object_type_from_user is None:
                raise RuntimeError(
                    f"Cannot import class '{type_name_from_metadata}' to convert '{document}' to an object, "
                    f"it might be removed from your project. Provide an alternative object type "
                    f"to convert the document to or pass 'dict' to receive dictionary JSON representation."
                )
            return object_type_from_user, False, False

        # Successfully imported the class from metadata - but before conversion check for projections and inheritance

        # Maybe user wants to cast from dict to their type
        if object_type_from_metadata is dict:
            return object_type_from_user, False, False

        # User doesn't need projection, or class from metadata is a child of a class provided by user
        # We can safely use class from metadata
        if object_type_from_user is None or Utils.is_inherit(object_type_from_user, object_type_from_metadata):
            return object_type_from_metadata, False, False

        # Passed type is not a type from metadata, neither there's no inheritance - probably projection
        elif object_type_from_user is not object_type_from_metadata:
            if not all([name in object_type_from_metadata.__dict__ for name in object_type_from_user.__dict__]):
                # Document from database and object_type from user aren't compatible
                raise exceptions.InvalidOperationException(
                    f"Cannot covert document from type {object_type_from_metadata} to {object_type_from_user}"
                )

            # Projection
            return object_type_from_user, True, False

    @staticmethod
    def set_python_type_in_metadata(metadata: Dict[str, Any], object_type: Type[Any]) -> None:
        metadata["Raven-Python-Type"] = "{0}.{1}".format(object_type.__module__, object_type.__name__)



from __future__ import annotations

import json
from datetime import timedelta
from typing import Dict, Optional, List

import requests

from ravendb.documents.subscriptions.state import SubscriptionState
from ravendb.http.topology import RaftCommand, ServerNode
from ravendb.documents.conventions import DocumentConventions
from ravendb.documents.subscriptions.options import SubscriptionCreationOptions, SubscriptionUpdateOptions
from ravendb.http.raven_command import RavenCommand, VoidRavenCommand
from ravendb.tools.utils import Utils
from ravendb.util.util import RaftIdGenerator


class CreateSubscriptionResult:
    def __init__(self, name: str):
        self.name = name

    @classmethod
    def from_json(cls, json_dict: Dict) -> CreateSubscriptionResult:
        return CreateSubscriptionResult(json_dict["Name"])


class CreateSubscriptionCommand(RavenCommand[CreateSubscriptionResult], RaftCommand):
    def __init__(
        self, conventions: DocumentConventions, options: SubscriptionCreationOptions, key: Optional[str] = None
    ):
        super().__init__(CreateSubscriptionResult)
        if options is None:
            raise ValueError("Options cannot be None")
        self._options = options
        self._key = key

    def create_request(self, node: ServerNode) -> requests.Request:
        url = f"{node.url}/databases/{node.database}/subscriptions"

        if self._key is not None:
            url += "?id" + Utils.quote_key(self._key)

        request = requests.Request("PUT")
        request.data = self._options.to_json()
        request.url = url
        return request

    def set_response(self, response: Optional[str], from_cache: bool) -> None:
        self.result = CreateSubscriptionResult.from_json(json.loads(response))

    def is_read_request(self) -> bool:
        return False

    def get_raft_unique_request_id(self) -> str:
        return RaftIdGenerator.new_id()


class TcpConnectionInfo:
    def __init__(
        self,
        port: Optional[int] = None,
        url: Optional[str] = None,
        certificate: Optional[str] = None,
        urls: Optional[List[str]] = None,
        node_tag: Optional[str] = None,
        server_id: Optional[str] = None,
    ):
        self.port = port
        self.url = url
        self.certificate = certificate
        self.urls = urls
        self.node_tag = node_tag
        self.server_id = server_id

    @classmethod
    def from_json(cls, json_dict: Dict) -> TcpConnectionInfo:
        return TcpConnectionInfo(
            json_dict.get("Port", None) or 0,
            json_dict.get("Url", None),
            json_dict.get("Certificate", None),
            json_dict.get("Urls", None),
            json_dict.get("NodeTag", None),
            json_dict.get("ServerId", None),
        )


class GetTcpInfoForRemoteTaskCommand(RavenCommand[TcpConnectionInfo]):
    def __init__(self, tag: str, remote_database: str, remote_task: str, verify_database: bool = False):
        super(GetTcpInfoForRemoteTaskCommand, self).__init__(TcpConnectionInfo)
        if remote_database is None:
            raise ValueError("remote_database cannot be None")

        self._remote_database = remote_database

        if remote_task is None:
            raise ValueError("remote_task cannot be None")

        self._remote_task = remote_task
        self._tag = tag
        self._verify_database = verify_database
        self.timeout = timedelta(seconds=15)
        self._requested_node: Optional[ServerNode] = None

    def create_request(self, node: ServerNode) -> requests.Request:
        url = (
            f"{node.url}/info/remote-task/tcp?"
            f"database={self._remote_database}"
            f"&remote-task={self._remote_task}"
            f"&tag={self._tag}"
        )

        if self._verify_database:
            url += "&verify-database=true"

        self._requested_node = node
        return requests.Request("GET", url)

    def set_response(self, response: Optional[str], from_cache: bool) -> None:
        if response is None:
            self._throw_invalid_response()

        self.result = TcpConnectionInfo.from_json(json.loads(response))

    @property
    def requested_node(self) -> ServerNode:
        return self._requested_node

    def is_read_request(self) -> bool:
        return True


class GetSubscriptionsResult:
    def __init__(self, results: Optional[List[SubscriptionState]] = None):
        self.results = results

    @classmethod
    def from_json(cls, json_dict: Dict) -> GetSubscriptionsResult:
        return cls([SubscriptionState.from_json(result) for result in json_dict["Results"]])


class GetSubscriptionsCommand(RavenCommand[List[SubscriptionState]]):
    def __init__(self, start: int, page_size: int):
        super(GetSubscriptionsCommand, self).__init__(list)
        self._start = start
        self._page_size = page_size

    def create_request(self, node: ServerNode) -> requests.Request:
        url = f"{node.url}/databases/{node.database}/subscriptions?start={self._start}&pageSize={self._page_size}"

        return requests.Request("GET", url)

    def set_response(self, response: Optional[str], from_cache: bool) -> None:
        if response is None:
            self.result = None
            return

        self.result = GetSubscriptionsResult.from_json(json.loads(response)).results

    def is_read_request(self) -> bool:
        return True


class DeleteSubscriptionCommand(VoidRavenCommand, RaftCommand):
    def __init__(self, name: str):
        super(DeleteSubscriptionCommand, self).__init__()
        self._name = name

    def create_request(self, node: ServerNode) -> requests.Request:
        url = f"{node.url}/databases/{node.database}/subscriptions?taskName={self._name}"

        return requests.Request("DELETE", url)

    def get_raft_unique_request_id(self) -> str:
        return RaftIdGenerator.new_id()


class GetSubscriptionStateCommand(RavenCommand[SubscriptionState]):
    def __init__(self, subscription_name: str):
        super(GetSubscriptionStateCommand, self).__init__(SubscriptionState)
        self._subscription_name = subscription_name

    def is_read_request(self) -> bool:
        return True

    def create_request(self, node: ServerNode) -> requests.Request:
        url = (
            f"{node.url}/databases/{node.database}/subscriptions/state?name={Utils.quote_key(self._subscription_name)}"
        )

        return requests.Request("GET", url)

    def set_response(self, response: Optional[str], from_cache: bool) -> None:
        self.result = SubscriptionState.from_json(json.loads(response))


class DropSubscriptionConnectionCommand(VoidRavenCommand):
    def __init__(self, name: Optional[str] = None):
        super(DropSubscriptionConnectionCommand, self).__init__()
        self._name = name

    def create_request(self, node: ServerNode) -> requests.Request:
        path = []
        path.append(node.url)
        path.append("/databases/")
        path.append(node.database)
        path.append("/subscriptions/drop")

        if self._name and not self._name.isspace():
            path.append("?name=")
            path.append(Utils.quote_key(self._name))

        url = "".join(path)
        return requests.Request("POST", url)


class UpdateSubscriptionResult(CreateSubscriptionResult):
    pass


class UpdateSubscriptionCommand(RavenCommand[UpdateSubscriptionResult], RaftCommand):
    def __init__(self, options: SubscriptionUpdateOptions):
        super(UpdateSubscriptionCommand, self).__init__(UpdateSubscriptionResult)
        self._options = options

    def create_request(self, node: ServerNode) -> requests.Request:
        url = f"{node.url}/databases/{node.database}/subscriptions/update"

        request = requests.Request("POST", url)
        request.data = self._options.to_json()

        return request

    def set_response(self, response: Optional[str], from_cache: bool) -> None:
        if from_cache:
            self.result = UpdateSubscriptionResult(self._options.name)
            return

        if response is None:
            self._throw_invalid_response()

        self.result = UpdateSubscriptionResult.from_json(json.loads(response))

    def is_read_request(self) -> bool:
        return False

    def get_raft_unique_request_id(self) -> str:
        return RaftIdGenerator.new_id()



import json
from typing import Optional, Union, TypeVar, Dict

import requests

from ravendb.documents.conventions import DocumentConventions
from ravendb.http.topology import RaftCommand
from ravendb.http.server_node import ServerNode
from ravendb.http.misc import Broadcast
from ravendb.http.raven_command import RavenCommand
from ravendb.documents.operations.definitions import MaintenanceOperation
from ravendb.util.util import RaftIdGenerator

_T_Result = TypeVar("_T_Result")


class NextIdentityForCommand(RavenCommand, RaftCommand, Broadcast):
    def __init__(self, _id: str):
        super().__init__(int)
        if _id is None:
            raise ValueError("Id cannot be None")

        self._id = _id
        self._raft_unique_request_id = RaftIdGenerator.new_id()

    @classmethod
    def from_copy(cls, copy: RavenCommand[_T_Result]) -> Union[Broadcast, RavenCommand[_T_Result]]:
        copied = super().from_copy(copy)

        copied._raft_unique_request_id = copy._raft_unique_request_id
        copied.id_ = copy._id
        return copied

    def is_read_request(self) -> bool:
        return False

    def create_request(self, node: ServerNode) -> requests.Request:
        self.ensure_is_not_null_or_string(self._id, "id")

        return requests.Request("POST", f"{node.url}/databases/{node.database}/identity/next?name={self._id}")

    def set_response(self, response: Optional[str], from_cache: bool) -> None:
        if response is None:
            self._throw_invalid_response()

        json_node = json.loads(response)
        if "NewIdentityValue" not in json_node:
            self._throw_invalid_response()

        self.result = json_node["NewIdentityValue"]

    def get_raft_unique_request_id(self) -> str:
        return self._raft_unique_request_id

    def prepare_to_broadcast(self, conventions: DocumentConventions) -> Broadcast:
        return NextIdentityForCommand.from_copy(self)


class NextIdentityForOperation(MaintenanceOperation[int]):
    def __init__(self, name: str):
        if not name or name.isspace():
            raise ValueError("The field name cannot be None or whitespace.")
        self._identity_name = name

    def get_command(self, conventions: "DocumentConventions") -> "RavenCommand[int]":
        return NextIdentityForCommand(self._identity_name)


class SeedIdentityForCommand(RavenCommand[int], RaftCommand):
    def __init__(self, id_: str, value: int = None, forced: bool = False):
        super().__init__(int)
        if id_ is None:
            raise ValueError("Id cannot be None")

        self._id = id_
        self._value = value
        self._forced = forced

    def is_read_request(self) -> bool:
        return False

    def create_request(self, node: ServerNode) -> requests.Request:
        self.ensure_is_not_null_or_string(self._id, "id")

        url = f"{node.url}/databases/{node.database}/identity/seed?name={self._id}&value={self._value}"

        if self._forced:
            url += "&force=true"

        return requests.Request("POST", url)

    def set_response(self, response: Optional[str], from_cache: bool) -> None:
        if response is None:
            self._throw_invalid_response()

        json_node = json.loads(response)

        if "NewSeedValue" not in json_node:
            self._throw_invalid_response()

        self.result = json_node["NewSeedValue"]

    def get_raft_unique_request_id(self) -> str:
        return RaftIdGenerator.new_id()


class SeedIdentityForOperation(MaintenanceOperation[int]):
    def __init__(self, name: str, value: int, force_update: bool = False):
        super().__init__()
        if not name or name.isspace():
            raise ValueError("The field name cannot be None or whitespace")

        self._identity_name = name
        self._identity_value = value
        self._force_update = force_update

    def get_command(self, conventions: "DocumentConventions") -> "RavenCommand[_T]":
        return SeedIdentityForCommand(self._identity_name, self._identity_value, self._force_update)


class GetIdentitiesOperation(MaintenanceOperation[Dict[str, int]]):
    def get_command(self, conventions: "DocumentConventions") -> "RavenCommand[Dict[str, int]]":
        return self.GetIdentitiesCommand()

    class GetIdentitiesCommand(RavenCommand[Dict[str, int]]):
        def __init__(self):
            super().__init__(dict)

        def is_read_request(self) -> bool:
            return True

        def create_request(self, node: ServerNode) -> requests.Request:
            url = f"{node.url}/databases/{node.database}/debug/identities"
            return requests.Request(method="GET", url=url)

        def set_response(self, response: Optional[str], from_cache: bool) -> None:
            self.result = json.loads(response)



from __future__ import annotations
import json
from datetime import timedelta
from enum import Enum
from typing import Optional, Any, Dict, TYPE_CHECKING

import requests

from ravendb import ServerNode
from ravendb.http.raven_command import RavenCommand, VoidRavenCommand
from ravendb.serverwide.operations.common import ServerOperation, T, VoidServerOperation
from ravendb.tools.utils import Utils

if TYPE_CHECKING:
    from ravendb.documents.conventions import DocumentConventions


class LogMode(Enum):
    NONE = "None"
    OPERATIONS = "Operations"
    INFORMATION = "Information"


class GetLogsConfigurationResult:
    def __init__(
        self,
        current_mode: LogMode = None,
        mode: LogMode = None,
        path: str = None,
        use_utc_time: bool = None,
        retention_time: timedelta = None,
        retention_size: int = None,
        compress: bool = None,
    ):
        self.current_mode = current_mode
        self.mode = mode
        self.path = path
        self.use_utc_time = use_utc_time
        self.retention_time = retention_time
        self.retention_size = retention_size
        self.compress = compress

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> GetLogsConfigurationResult:
        return cls(
            LogMode(json_dict["CurrentMode"]),
            LogMode(json_dict["Mode"]),
            json_dict["Path"],
            json_dict["UseUtcTime"],
            Utils.string_to_timedelta(json_dict["RetentionTime"]),
            json_dict["RetentionSize"],
            json_dict["Compress"],
        )


class GetLogsConfigurationOperation(ServerOperation[GetLogsConfigurationResult]):
    def get_command(self, conventions: "DocumentConventions") -> RavenCommand[T]:
        return self.GetLogsConfigurationCommand()

    class GetLogsConfigurationCommand(RavenCommand[GetLogsConfigurationResult]):
        def __init__(self):
            super().__init__(GetLogsConfigurationResult)

        def is_read_request(self) -> bool:
            return False

        def create_request(self, node: ServerNode) -> requests.Request:
            return requests.Request("GET", f"{node.url}/admin/logs/configuration")

        def set_response(self, response: Optional[str], from_cache: bool) -> None:
            if response is None:
                self._throw_invalid_response()

            self.result = GetLogsConfigurationResult.from_json(json.loads(response))


class SetLogsConfigurationOperation(VoidServerOperation):
    class Parameters:
        def __init__(
            self,
            mode: LogMode = None,
            retention_time: timedelta = None,
            retention_size: int = None,
            compress: bool = None,
        ) -> None:
            self.mode = mode
            self.retention_time = retention_time
            self.retention_size = retention_size
            self.compress = compress

        @classmethod
        def from_get_logs(cls, get_logs: GetLogsConfigurationResult) -> "SetLogsConfigurationOperation.Parameters":
            return cls(get_logs.mode, get_logs.retention_time, get_logs.retention_size, get_logs.compress)

        def to_json(self) -> Dict[str, Any]:
            return {
                "Mode": self.mode.value,
                "RetentionTime": Utils.timedelta_to_str(self.retention_time),
                "RetentionSize": self.retention_size,
                "Compress": self.compress,
            }

    def __init__(self, parameters: Parameters) -> None:
        if parameters is None:
            raise ValueError("Parameters cannot be None")

        self._parameters = parameters

    def get_command(self, conventions: "DocumentConventions") -> "VoidRavenCommand":
        return self.SetLogsConfigurationCommand(self._parameters)

    class SetLogsConfigurationCommand(VoidRavenCommand):
        def __init__(self, parameters: "SetLogsConfigurationOperation.Parameters"):
            if parameters is None:
                raise ValueError("Parameters must be None")

            super().__init__()
            self._parameters = parameters

        def create_request(self, node: ServerNode) -> requests.Request:
            url = f"{node.url}/admin/logs/configuration"

            request = requests.Request("POST", url, data=self._parameters.to_json())
            return request



from __future__ import annotations

import json

from ravendb.exceptions.raven_exceptions import ConflictException, BadResponseException


class DocumentConflictException(ConflictException):
    def __init__(self, message: str, doc_id: str = None, largest_etag: int = None):
        super(DocumentConflictException, self).__init__(message)
        self.doc_id = doc_id
        self.largest_etag = largest_etag

    @classmethod
    def from_message(cls, message: str) -> DocumentConflictException:
        return cls(message, None, 0)

    @classmethod
    def from_json(cls, json_s: str) -> DocumentConflictException:
        try:
            json_node = json.loads(json_s)
            doc_id = json_node.get("DocId")
            message = json_node.get("Message")
            largest_etag = json_node.get("LargestEtag")

            return cls(str(message), str(doc_id), int(largest_etag))
        except IOError as e:
            raise BadResponseException("Unable to parse server response: ", e)



from __future__ import annotations
import datetime
from enum import Enum
from typing import Union, Optional, Generic, TypeVar, Type

from ravendb.primitives import constants
from ravendb.documents.conventions import DocumentConventions
from ravendb.documents.commands.batches import PutCompareExchangeCommandData, DeleteCompareExchangeCommandData
from ravendb.documents.session.document_session_operations.misc import _update_metadata_modifications
from ravendb.exceptions.raven_exceptions import RavenException
from ravendb.json.metadata_as_dictionary import MetadataAsDictionary
from ravendb.documents.session.entity_to_json import EntityToJsonStatic
from ravendb.tools.utils import Utils

_T = TypeVar("_T")


class CompareExchangeValueState(Enum):
    NONE = "none"
    CREATED = "created"
    DELETED = "deleted"
    MISSING = "missing"

    def __str__(self):
        return self.value


class CompareExchangeValue(Generic[_T]):
    def __init__(self, key: str, index: int, value: _T, metadata: MetadataAsDictionary = None):
        self.key = key
        self.index = index
        self.value = value
        self._metadata_as_dictionary = metadata

    @property
    def metadata(self) -> MetadataAsDictionary:
        if not self._metadata_as_dictionary:
            self._metadata_as_dictionary = MetadataAsDictionary()
        return self._metadata_as_dictionary

    @property
    def has_metadata(self) -> bool:
        return self._metadata_as_dictionary is not None


class CompareExchangeSessionValue(Generic[_T]):
    def __init__(
        self,
        key: str = None,
        index: int = None,
        state: CompareExchangeValueState = None,
        value: Optional[CompareExchangeValue[_T]] = None,
    ):
        if not ((value is not None) ^ all([key is not None, index is not None, state is not None])):
            raise ValueError("Specify key, index and state or just pass CompareExchangeValue.")
        if value:
            self.__init__(
                value.key,
                value.index,
                CompareExchangeValueState.NONE if value.index >= 0 else CompareExchangeValueState.MISSING,
            )
            if value.index > 0:
                self.__original_value = value
            return
        self._key = key
        self._index = index
        self._state = state
        self.__original_value: Union[CompareExchangeValue, None] = None
        self.__value: Union[CompareExchangeValue, None] = value

    def get_value(
        self, object_type: Optional[Type[_T]], conventions: DocumentConventions
    ) -> Optional[CompareExchangeValue[_T]]:
        if (
            self._state.value == CompareExchangeValueState.CREATED.value
            or self._state.value == CompareExchangeValueState.NONE.value
        ):
            if isinstance(self.__value, CompareExchangeValue):
                return self.__value

            if self.__value:
                raise ValueError("Value cannot be None")  # todo: RavenDB-17249

            entity = None
            if self.__original_value is not None and self.__original_value.value is not None:
                if object_type in [int, float, str, bool, bytes]:
                    try:
                        entity = self.__original_value.value
                    except BaseException as ex:
                        raise RavenException(
                            f"Unable to read compare exchange value: {self.__original_value.value}", ex
                        )
                elif isinstance(self.__original_value.value, dict):
                    entity = Utils.convert_json_dict_to_object(self.__original_value.value, object_type)
                else:
                    entity = self.__original_value.value

                value = CompareExchangeValue(self._key, self._index, entity)
                self.__value = value
                return value

            value = CompareExchangeValue(
                self._key,
                self._index,
                entity,
                self.__original_value.metadata if self.__original_value is not None else None,
            )
            self.__value = value

            return value

        elif (
            self._state.value == CompareExchangeValueState.MISSING.value
            or self._state.value == CompareExchangeValueState.DELETED.value
        ):
            return None
        else:
            raise ValueError(f"Not supported state: {self._state}")

    def create(self, item) -> CompareExchangeValue[_T]:
        self.__assert_state()
        if self.__value:
            raise RuntimeError(f"The compare exchange value with key {self._key} is already tracked.")
        self._index = 0
        value = CompareExchangeValue(self._key, self._index, item)
        self.__value = value
        self._state = CompareExchangeValueState.CREATED
        return value

    def delete(self, index: int) -> None:
        self.__assert_state()
        self._index = index
        self._state = CompareExchangeValueState.DELETED

    def __assert_state(self) -> None:
        if self._state == CompareExchangeValueState.NONE or CompareExchangeValueState.MISSING:
            return
        elif self._state == CompareExchangeValueState.CREATED:
            raise RuntimeError(f"The compare exchange value with key {self._key} was already stored.")
        elif self._state == CompareExchangeValueState.DELETED:
            raise RuntimeError(f"The compare exchange value with key {self._key} was already deleted")

    def get_command(
        self, conventions: DocumentConventions
    ) -> Optional[Union[DeleteCompareExchangeCommandData, PutCompareExchangeCommandData]]:
        s = self._state
        if s == CompareExchangeValueState.NONE or CompareExchangeValueState.CREATED:
            if not self.__value:
                return None

            entity = EntityToJsonStatic.convert_entity_to_json(self.__value.value, conventions, None, False)

            entity_json = entity if isinstance(entity, dict) else None
            metadata = (
                self.__original_value.value.get(constants.Documents.Metadata.KEY, None)
                if self.__original_value is not None
                else None
            )
            metadata_has_changed = False

            if self.__value.has_metadata and len(self.__value.metadata) != 0:
                if metadata is None:
                    metadata_has_changed = True
                    # create new metadata (because there wasn't any metadata before)
                    metadata = self.prepare_metadata_for_put(self._key, self.__value.metadata, conventions)
                else:
                    self.validate_metadata_for_put(self._key, self.__value.metadata)
                    metadata_has_changed = _update_metadata_modifications(self.__value.metadata, metadata)

            entity_to_insert = None
            if not entity_json or metadata_has_changed:
                entity_json = entity_to_insert = self.__convert_entity(self._key, entity, metadata)

            new_value = CompareExchangeValue(self._key, self._index, entity_json)
            has_changed = (
                self.__original_value is None
                or metadata_has_changed
                or self.has_changed(self.__original_value, new_value)
            )
            self.__original_value = new_value
            if not has_changed:
                return None

            if entity_to_insert == None:
                entity_to_insert = self.__convert_entity(self._key, entity, metadata)

            return PutCompareExchangeCommandData(new_value.key, entity_to_insert, new_value.index)
        elif s == CompareExchangeValueState.DELETED:
            return DeleteCompareExchangeCommandData(self._key, self._index)
        elif s == CompareExchangeValueState.MISSING:
            return None
        else:
            raise ValueError(f"Not supported state: {self._state}")

    def __convert_entity(self, key: str, entity: _T, metadata) -> dict:
        object_node = {constants.CompareExchange.OBJECT_FIELD_NAME: Utils.entity_to_dict(entity, self)}
        if metadata:
            object_node[constants.Documents.Metadata.KEY] = metadata
        return object_node

    def has_changed(self, original_value: CompareExchangeValue, new_value: CompareExchangeValue):
        if original_value == new_value:
            return False

        if original_value.key.lower() != new_value.key.lower():
            raise ValueError(f"Keys do not match. Expected '{original_value.key}' but was: {new_value.key}'")

        if original_value.index != new_value.index:
            return True

        return original_value.value != new_value.value

    def update_state(self, index: int) -> None:
        self._index = index
        self._state = CompareExchangeValueState.NONE

        if self.__original_value is not None:
            self.__original_value.index = index

        if self.__value is not None:
            self.__value.index = index

    def update_value(self, value: CompareExchangeValue):
        self._index = value.index
        self._state = CompareExchangeValueState.NONE if value.index >= 0 else CompareExchangeValueState.MISSING
        self.__original_value = value

        if self.__value is not None:
            self.__value.index = self._index

            if self.__value.value is not None:
                EntityToJsonStatic.populate_entity(self.__value.value, value.value)

    @staticmethod
    def prepare_metadata_for_put(
        key: str, metadata_dictionary: MetadataAsDictionary, conventions: DocumentConventions
    ) -> dict:
        CompareExchangeSessionValue.validate_metadata_for_put(key, metadata_dictionary)

        return metadata_dictionary.metadata

    @staticmethod
    def validate_metadata_for_put(key: str, metadata_as_dictionary: MetadataAsDictionary) -> None:
        if constants.Documents.Metadata.EXPIRES in metadata_as_dictionary:
            obj = metadata_as_dictionary[constants.Documents.Metadata.EXPIRES]
            if obj is None:
                f"The values of {constants.Documents.Metadata.EXPIRES} metadata for compare exchange '{key}' is None"
            if not isinstance(obj, datetime.datetime) and not isinstance(obj, str):
                raise ValueError(
                    f"The type of {constants.Documents.Metadata.EXPIRES} metadata for compare exchange '{key}' is not valid. Use the following type: datetime.datetime or string"
                )



from __future__ import annotations

from datetime import datetime
from typing import Optional

from ravendb.primitives.exceptions import OperationCancelledException


class CancellationTokenSource:
    def __init__(self):
        self.cancelled = False
        self._cancel_after_date: Optional[float] = None

    def get_token(self) -> CancellationToken:
        return self.CancellationToken(self)

    class CancellationToken:
        def __init__(self, source: CancellationTokenSource):
            self.__source = source

        def is_cancellation_requested(self) -> bool:
            return self.__source.cancelled or (
                self.__source._cancel_after_date is not None
                and datetime.now().timestamp() > self.__source._cancel_after_date
            )

        def throw_if_cancellation_requested(self) -> None:
            if self.is_cancellation_requested():
                raise OperationCancelledException()

    def cancel(self) -> None:
        self.cancelled = True

    def cancel_after(self, timeout_in_seconds: float) -> None:
        self._cancel_after_date = datetime.now().timestamp() + timeout_in_seconds



from typing import Optional

from ravendb.documents.operations.backups.settings import (
    LocalSettings,
    S3Settings,
    AzureSettings,
    GlacierSettings,
    GoogleCloudSettings,
    FtpSettings,
)
from ravendb.documents.operations.connection_strings import ConnectionString
import ravendb.serverwide
from ravendb.documents.operations.etl.configuration import EtlConfiguration


class OlapConnectionString(ConnectionString):
    def __init__(
        self,
        name: str,
        local_settings: Optional[LocalSettings] = None,
        s3_settings: Optional[S3Settings] = None,
        azure_settings: Optional[AzureSettings] = None,
        glacier_settings: Optional[GlacierSettings] = None,
        google_cloud_settings: Optional[GoogleCloudSettings] = None,
        ftp_settings: Optional[FtpSettings] = None,
    ):
        super().__init__(name)
        self.local_settings = local_settings
        self.s3_settings = s3_settings
        self.azure_settings = azure_settings
        self.glacier_settings = glacier_settings
        self.google_cloud_settings = google_cloud_settings
        self.ftp_settings = ftp_settings

    @property
    def get_type(self):
        return ravendb.serverwide.ConnectionStringType.OLAP


# todo: implement
class OlapEtlConfiguration(EtlConfiguration[OlapConnectionString]):
    pass



from __future__ import annotations
from typing import Union, Optional, List, Dict, Any


class GetDocumentResult:
    def __init__(self):
        self.includes: Union[None, dict] = None
        self.results: Union[None, list] = None
        self.counter_includes: Union[None, dict] = None
        self.time_series_includes: Union[None, dict] = None
        self.compare_exchange_value_includes: Union[None, dict] = None
        self.next_page_start: Union[None, int] = None


class GetDocumentsResult:
    def __init__(
        self,
        includes: Optional[dict] = None,
        results: Optional[List[Dict]] = None,
        counter_includes: Optional[dict] = None,
        time_series_includes: Optional[dict] = None,
        compare_exchange_includes: Optional[dict] = None,
        next_page_start: Optional[int] = None,
    ):
        self.includes = includes
        self.results = results
        self.counter_includes = counter_includes
        self.time_series_includes = time_series_includes
        self.compare_exchange_includes = compare_exchange_includes
        self.next_page_start = next_page_start

    @classmethod
    def from_json(cls, json_dict: dict) -> GetDocumentsResult:
        return cls(
            json_dict.get("Includes", None),
            json_dict.get("Results", None),
            json_dict.get("CounterIncludes", None),
            json_dict.get("TimeSeriesIncludes", None),
            json_dict.get("CompareExchangeValueIncludes", None),
            json_dict.get("NextPageStart", None),
        )

    def to_json(self) -> Dict[str, Any]:
        return {
            "Includes": self.includes,
            "Results": self.results,
            "CounterIncludes": self.counter_includes,
            "TimeSeriesIncludes": self.time_series_includes,
            "CompareExchangeValueIncludes": self.compare_exchange_includes,
            "NextPageStart": self.next_page_start,
        }
